# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""ETCI 2021 dataset."""

import glob
import os
from collections.abc import Callable
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_and_extract_archive


class ETCI2021(NonGeoDataset):
    """ETCI 2021 Flood Detection dataset.

    The `ETCI2021 <https://nasa-impact.github.io/etci2021/>`_
    dataset is a dataset for flood detection

    Dataset features:

    * 33,405 VV & VH Sentinel-1 Synthetic Aperture Radar (SAR) images
    * 2 binary masks per image representing water body & flood, respectively
    * 2 polarization band images (VV, VH) of 3 RGB channels per band
    * 3 RGB channels per band generated by the Hybrid Pluggable Processing
      Pipeline (hyp3)
    * Images with 5x20m per pixel resolution (256x256) px) taken in
      Interferometric Wide Swath acquisition mode
    * Flood events from 5 different regions

    Dataset format:

    * VV band three-channel png
    * VH band three-channel png
    * water body mask single-channel png where no water body = 0, water body = 255
    * flood mask single-channel png where no flood = 0, flood = 255

    Dataset classes:

    1. no flood/water
    2. flood/water

    If you use this dataset in your research, please add the following to your
    acknowledgements section::

        The authors would like to thank the NASA Earth Science Data Systems Program,
        NASA Digital Transformation AI/ML thrust, and IEEE GRSS for organizing
        the ETCI competition.
    """

    bands = ('VV', 'VH')
    masks = ('flood', 'water_body')
    metadata: ClassVar[dict[str, dict[str, str]]] = {
        'train': {
            'filename': 'train.zip',
            'md5': '1e95792fe0f6e3c9000abdeab2a8ab0f',
            'directory': 'train',
            'url': 'https://drive.google.com/file/d/14HqNW5uWLS92n7KrxKgDwUTsSEST6LCr',
        },
        'val': {
            'filename': 'val_with_ref_labels.zip',
            'md5': 'fd18cecb318efc69f8319f90c3771bdf',
            'directory': 'test',
            'url': 'https://drive.google.com/file/d/19sriKPHCZLfJn_Jmk3Z_0b3VaCBVRVyn',
        },
        'test': {
            'filename': 'test_without_ref_labels.zip',
            'md5': 'da9fa69e1498bd49d5c766338c6dac3d',
            'directory': 'test_internal',
            'url': 'https://drive.google.com/file/d/1rpMVluASnSHBfm2FhpPDio0GyCPOqg7E',
        },
    }

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new ETCI 2021 dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` argument is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.metadata.keys()

        self.root = root
        self.split = split
        self.transforms = transforms
        self.checksum = checksum

        if download:
            self._download()

        if not self._check_integrity():
            raise DatasetNotFoundError(self)

        self.files = self._load_files(self.root, self.split)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        files = self.files[index]
        vv = self._load_image(files['vv'])
        vh = self._load_image(files['vh'])
        water_mask = self._load_target(files['water_mask'])

        if self.split != 'test':
            flood_mask = self._load_target(files['flood_mask'])
            mask = torch.stack(tensors=[water_mask, flood_mask], dim=0)
        else:
            mask = water_mask.unsqueeze(0)

        image = torch.cat(tensors=[vv, vh], dim=0)
        sample = {'image': image, 'mask': mask}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def _load_files(self, root: Path, split: str) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Args:
            root: root dir of dataset
            split: subset of dataset, one of [train, val, test]

        Returns:
            list of dicts containing paths for each pair of vv, vh,
            water body mask, flood mask (train/val only)
        """
        files = []
        directory = self.metadata[split]['directory']
        folders = sorted(glob.glob(os.path.join(root, directory, '*')))
        folders = [os.path.join(folder, 'tiles') for folder in folders]
        for folder in folders:
            vvs = sorted(glob.glob(os.path.join(folder, 'vv', '*.png')))
            vhs = sorted(glob.glob(os.path.join(folder, 'vh', '*.png')))
            water_masks = sorted(
                glob.glob(os.path.join(folder, 'water_body_label', '*.png'))
            )

            if split != 'test':
                flood_masks = sorted(
                    glob.glob(os.path.join(folder, 'flood_label', '*.png'))
                )

                for vv, vh, flood_mask, water_mask in zip(
                    vvs, vhs, flood_masks, water_masks
                ):
                    files.append(
                        dict(vv=vv, vh=vh, flood_mask=flood_mask, water_mask=water_mask)
                    )
            else:
                for vv, vh, water_mask in zip(vvs, vhs, water_masks):
                    files.append(dict(vv=vv, vh=vh, water_mask=water_mask))

        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the image
        """
        filename = os.path.join(path)
        with Image.open(filename) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('RGB'))
            tensor = torch.from_numpy(array).float()
            # Convert from HxWxC to CxHxW
            tensor = tensor.permute((2, 0, 1))
            return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Load the target mask for a single image.

        Args:
            path: path to the image

        Returns:
            the target mask
        """
        filename = os.path.join(path)
        with Image.open(filename) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('L'))
            tensor = torch.from_numpy(array)
            tensor = torch.clamp(tensor, min=0, max=1)
            tensor = tensor.to(torch.long)
            return tensor

    def _check_integrity(self) -> bool:
        """Checks the integrity of the dataset structure.

        Returns:
            True if the dataset directories and split files are found, else False
        """
        directory = self.metadata[self.split]['directory']
        dirpath = os.path.join(self.root, directory)
        if not os.path.exists(dirpath):
            return False
        return True

    def _download(self) -> None:
        """Download the dataset and extract it."""
        if self._check_integrity():
            print('Files already downloaded and verified')
            return

        download_and_extract_archive(
            self.metadata[self.split]['url'],
            self.root,
            filename=self.metadata[self.split]['filename'],
            md5=self.metadata[self.split]['md5'] if self.checksum else None,
        )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        vv = np.rollaxis(sample['image'][:3].numpy(), 0, 3)
        vh = np.rollaxis(sample['image'][3:].numpy(), 0, 3)
        mask = sample['mask'].squeeze(0)

        showing_flood_mask = mask.shape[0] == 2
        showing_predictions = 'prediction' in sample
        num_panels = 3
        if showing_flood_mask:
            water_mask = mask[0].numpy()
            flood_mask = mask[1].numpy()
            num_panels += 1
        else:
            water_mask = mask.numpy()

        if showing_predictions:
            predictions = sample['prediction'].numpy()
            num_panels += 1

        fig, axs = plt.subplots(1, num_panels, figsize=(num_panels * 4, 3))
        axs[0].imshow(vv)
        axs[0].axis('off')
        axs[1].imshow(vh)
        axs[1].axis('off')
        axs[2].imshow(water_mask)
        axs[2].axis('off')
        if show_titles:
            axs[0].set_title('VV')
            axs[1].set_title('VH')
            axs[2].set_title('Water mask')

        idx = 0
        if showing_flood_mask:
            axs[3 + idx].imshow(flood_mask)
            axs[3 + idx].axis('off')
            if show_titles:
                axs[3 + idx].set_title('Flood mask')
            idx += 1

        if showing_predictions:
            axs[3 + idx].imshow(predictions)
            axs[3 + idx].axis('off')
            if show_titles:
                axs[3 + idx].set_title('Predictions')
            idx += 1

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Simple fully convolutional neural network (FCN) implementations."""

import torch.nn as nn
from torch import Tensor
from torch.nn.modules import Module


class FCN(Module):
    """A simple 5 layer FCN with leaky relus and 'same' padding."""

    def __init__(self, in_channels: int, classes: int, num_filters: int = 64) -> None:
        """Initializes the 5 layer FCN model.

        Args:
            in_channels: Number of input channels that the model will expect
            classes: Number of filters in the final layer
            num_filters: Number of filters in each convolutional layer
        """
        super().__init__()

        conv1 = nn.modules.Conv2d(
            in_channels, num_filters, kernel_size=3, stride=1, padding=1
        )
        conv2 = nn.modules.Conv2d(
            num_filters, num_filters, kernel_size=3, stride=1, padding=1
        )
        conv3 = nn.modules.Conv2d(
            num_filters, num_filters, kernel_size=3, stride=1, padding=1
        )
        conv4 = nn.modules.Conv2d(
            num_filters, num_filters, kernel_size=3, stride=1, padding=1
        )
        conv5 = nn.modules.Conv2d(
            num_filters, num_filters, kernel_size=3, stride=1, padding=1
        )

        self.backbone = nn.modules.Sequential(
            conv1,
            nn.modules.LeakyReLU(inplace=True),
            conv2,
            nn.modules.LeakyReLU(inplace=True),
            conv3,
            nn.modules.LeakyReLU(inplace=True),
            conv4,
            nn.modules.LeakyReLU(inplace=True),
            conv5,
            nn.modules.LeakyReLU(inplace=True),
        )

        self.last = nn.modules.Conv2d(
            num_filters, classes, kernel_size=1, stride=1, padding=0
        )

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass of the model."""
        x = self.backbone(x)
        x = self.last(x)
        return x



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 2 dataset."""

import os
from typing import ClassVar

from .base import SpaceNet


class SpaceNet2(SpaceNet):
    r"""SpaceNet 2: Building Detection v2 Dataset.

    `SpaceNet 2 <https://spacenet.ai/spacenet-buildings-dataset-v2/>`_
    is a dataset of building footprints over the cities of Las Vegas,
    Paris, Shanghai and Khartoum.

    Collection features:

    +------------+---------------------+------------+------------+
    |    AOI     | Area (km\ :sup:`2`\)| # Images   | # Buildings|
    +============+=====================+============+============+
    | Las Vegas  |    216              |   3850     |  151,367   |
    +------------+---------------------+------------+------------+
    | Paris      |    1030             |   1148     |  23,816    |
    +------------+---------------------+------------+------------+
    | Shanghai   |    1000             |   4582     |  92,015    |
    +------------+---------------------+------------+------------+
    | Khartoum   |    765              |   1012     |  35,503    |
    +------------+---------------------+------------+------------+

    Imagery features:

    .. list-table::
        :widths: 10 10 10 10 10
        :header-rows: 1
        :stub-columns: 1

        *   -
            - PAN
            - MS
            - PS-MS
            - PS-RGB
        *   - GSD (m)
            - 0.31
            - 1.24
            - 0.30
            - 0.30
        *   - Chip size (px)
            - 650 x 650
            - 163 x 163
            - 650 x 650
            - 650 x 650

    Dataset format:

    * Imagery - Worldview-3 GeoTIFFs

        * PAN.tif (Panchromatic)
        * MS.tif (Multispectral)
        * PS-MS (Pansharpened Multispectral)
        * PS-RGB (Pansharpened RGB)

    * Labels - GeoJSON

        * label.geojson

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/1807.01232
    """

    dataset_id = 'SN2_buildings'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            2: ['SN2_buildings_train_AOI_2_Vegas.tar.gz'],
            3: ['SN2_buildings_train_AOI_3_Paris.tar.gz'],
            4: ['SN2_buildings_train_AOI_4_Shanghai.tar.gz'],
            5: ['SN2_buildings_train_AOI_5_Khartoum.tar.gz'],
        },
        'test': {
            2: ['AOI_2_Vegas_Test_public.tar.gz'],
            3: ['AOI_3_Paris_Test_public.tar.gz'],
            4: ['AOI_4_Shanghai_Test_public.tar.gz'],
            5: ['AOI_5_Khartoum_Test_public.tar.gz'],
        },
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            2: ['307da318bc43aaf9481828f92eda9126'],
            3: ['4db469e3e4e7bf025368ad730aec0888'],
            4: ['986129eecd3e842ebc2063d43b407adb'],
            5: ['462b4bf0466c945d708befabd4d9115b'],
        },
        'test': {
            2: ['d45405afd6629e693e2f9168b1291ea3'],
            3: ['2eaee95303e88479246e4ee2f2279b7f'],
            4: ['f51dc51fa484dc7fb89b3697bd15a950'],
            5: ['037d7be10530f0dd1c43d4ef79f3236e'],
        },
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {
        'train': [2, 3, 4, 5],
        'test': [2, 3, 4, 5],
    }
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['MUL', 'MUL-PanSharpen', 'PAN', 'RGB-PanSharpen'],
        'test': ['MUL', 'MUL-PanSharpen', 'PAN', 'RGB-PanSharpen'],
    }
    valid_masks = (os.path.join('geojson', 'buildings'),)
    chip_size: ClassVar[dict[str, tuple[int, int]]] = {'MUL': (163, 163)}



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TorchGeo: datasets, samplers, transforms, and pre-trained models for geospatial data.

This library is part of the `PyTorch <https://pytorch.org/>`_ project. PyTorch is an
open source machine learning framework.

The :mod:`torchgeo` package consists of popular datasets, model architectures, and
common image transformations for geospatial data.
"""

__author__ = 'Adam J. Stewart'
__version__ = '0.7.1'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""GlobalBuildingMap."""

from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from .geo import RasterDataset


class GlobalBuildingMap(RasterDataset):
    """GlobalBuildingMap dataset.

    The GlobalBuildingMap (GBM) dataset provides the highest resolution and highest
    accuracy building footprint map on a global scale ever created. GBM was generated
    by training and applying modern deep neural networks on nearly 800,000 satellite
    images. The dataset is stored in 5 by 5 degree tiles in geotiff format.

    The GlobalBuildingMap is generated by applying an ensemble of deep neural networks
    on nearly 800,000 satellite images of about 3m resolution. The deep neural networks
    were trained with manually inspected training samples generated from OpenStreetMap.

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2404.13911

    .. versionadded:: 0.7
    """

    filename_glob = 'GBM_v1_*'
    is_image = False

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: A sample returned by :meth:`RasterDataset.__getitem__`.
            show_titles: Flag indicating whether to show titles above each panel.
            suptitle: Optional string to use as a suptitle.

        Returns:
            A matplotlib Figure with the rendered sample.
        """
        ncols = 2 if 'prediction' in sample else 1
        fig, axs = plt.subplots(ncols=ncols, squeeze=False)

        axs[0, 0].imshow(sample['mask'], cmap='gray')
        axs[0, 0].axis('off')
        if show_titles:
            axs[0, 0].set_title('Mask')

        if 'prediction' in sample:
            axs[0, 1].imshow(sample['prediction'], cmap='gray')
            axs[0, 1].axis('off')
            if show_titles:
                axs[0, 1].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench datasets."""

from typing import Any, Literal

from torch import Tensor

from ..geo import NonGeoDataset
from .aq_no2_s5p import CopernicusBenchAQNO2S5P
from .aq_o3_s5p import CopernicusBenchAQO3S5P
from .base import CopernicusBenchBase
from .bigearthnet_s1 import CopernicusBenchBigEarthNetS1
from .bigearthnet_s2 import CopernicusBenchBigEarthNetS2
from .biomass_s3 import CopernicusBenchBiomassS3
from .cloud_s2 import CopernicusBenchCloudS2
from .cloud_s3 import CopernicusBenchCloudS3
from .dfc2020_s1 import CopernicusBenchDFC2020S1
from .dfc2020_s2 import CopernicusBenchDFC2020S2
from .eurosat_s1 import CopernicusBenchEuroSATS1
from .eurosat_s2 import CopernicusBenchEuroSATS2
from .flood_s1 import CopernicusBenchFloodS1
from .lc100cls_s3 import CopernicusBenchLC100ClsS3
from .lc100seg_s3 import CopernicusBenchLC100SegS3
from .lcz_s2 import CopernicusBenchLCZS2
from .pretrain import CopernicusPretrain

__all__ = (
    'CopernicusBench',
    'CopernicusBenchAQNO2S5P',
    'CopernicusBenchAQO3S5P',
    'CopernicusBenchBase',
    'CopernicusBenchBigEarthNetS1',
    'CopernicusBenchBigEarthNetS2',
    'CopernicusBenchBiomassS3',
    'CopernicusBenchCloudS2',
    'CopernicusBenchCloudS3',
    'CopernicusBenchDFC2020S1',
    'CopernicusBenchDFC2020S2',
    'CopernicusBenchEuroSATS1',
    'CopernicusBenchEuroSATS2',
    'CopernicusBenchFloodS1',
    'CopernicusBenchLC100ClsS3',
    'CopernicusBenchLC100SegS3',
    'CopernicusBenchLCZS2',
    'CopernicusPretrain',
)

DATASET_REGISTRY = {
    'cloud_s2': CopernicusBenchCloudS2,
    'cloud_s3': CopernicusBenchCloudS3,
    'eurosat_s1': CopernicusBenchEuroSATS1,
    'eurosat_s2': CopernicusBenchEuroSATS2,
    'bigearthnet_s1': CopernicusBenchBigEarthNetS1,
    'bigearthnet_s2': CopernicusBenchBigEarthNetS2,
    'lc100cls_s3': CopernicusBenchLC100ClsS3,
    'lc100seg_s3': CopernicusBenchLC100SegS3,
    'dfc2020_s1': CopernicusBenchDFC2020S1,
    'dfc2020_s2': CopernicusBenchDFC2020S2,
    'flood_s1': CopernicusBenchFloodS1,
    'lcz_s2': CopernicusBenchLCZS2,
    'biomass_s3': CopernicusBenchBiomassS3,
    'aq_no2_s5p': CopernicusBenchAQNO2S5P,
    'aq_o3_s5p': CopernicusBenchAQO3S5P,
}


class CopernicusBench(NonGeoDataset):
    """Copernicus-Bench datasets.

    This wrapper supports dynamically loading datasets in Copernicus-Bench.

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2503.11849

    .. versionadded:: 0.7
    """

    def __init__(
        self,
        name: Literal[
            'cloud_s2',
            'cloud_s3',
            'eurosat_s1',
            'eurosat_s2',
            'bigearthnet_s1',
            'bigearthnet_s2',
            'lc100cls_s3',
            'lc100seg_s3',
            'dfc2020_s1',
            'dfc2020_s2',
            'flood_s1',
            'lcz_s2',
            'biomass_s3',
            'aq_no2_s5p',
            'aq_o3_s5p',
        ],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize a new CopernicusBench instance.

        Args:
            name: Name of the dataset to load.
            *args: Arguments to pass to dataset class.
            **kwargs: Keyword arguments to pass to dataset class.
        """
        self.name = name
        self.dataset: CopernicusBenchBase = DATASET_REGISTRY[name](*args, **kwargs)

    def __len__(self) -> int:
        """Return the length of the dataset.

        Returns:
            Length of the dataset.
        """
        return len(self.dataset)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        return self.dataset[index]

    def __getattr__(self, name: str) -> Any:
        """Wrapper around dataset object."""
        return getattr(self.dataset, name)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TorchGeo samplers."""

from .batch import BatchGeoSampler, RandomBatchGeoSampler
from .constants import Units
from .single import GeoSampler, GridGeoSampler, PreChippedGeoSampler, RandomGeoSampler
from .utils import get_random_bounding_box, tile_to_chips

__all__ = (
    'BatchGeoSampler',
    'GeoSampler',
    'GridGeoSampler',
    'PreChippedGeoSampler',
    'RandomBatchGeoSampler',
    'RandomGeoSampler',
    'Units',
    'get_random_bounding_box',
    'tile_to_chips',
)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""ETCI 2021 datamodule."""

from typing import Any

import torch
from torch import Tensor

from ..datasets import ETCI2021
from .geo import NonGeoDataModule


class ETCI2021DataModule(NonGeoDataModule):
    """LightningDataModule implementation for the ETCI2021 dataset.

    Splits the existing train split from the dataset into train/val with 80/20
    proportions, then uses the existing val dataset as the test data.

    .. versionadded:: 0.2
    """

    mean = torch.tensor(
        [
            128.02253931,
            128.02253931,
            128.02253931,
            128.11221701,
            128.11221701,
            128.11221701,
        ]
    )
    std = torch.tensor(
        [89.8145088, 89.8145088, 89.8145088, 95.2797861, 95.2797861, 95.2797861]
    )

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new ETCI2021DataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.ETCI2021`.
        """
        super().__init__(ETCI2021, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit']:
            self.train_dataset = ETCI2021(split='train', **self.kwargs)
        if stage in ['fit', 'validate']:
            self.val_dataset = ETCI2021(split='val', **self.kwargs)
        if stage in ['predict']:
            # Test set masks are not public, use for prediction instead
            self.predict_dataset = ETCI2021(split='test', **self.kwargs)

    def on_after_batch_transfer(
        self, batch: dict[str, Tensor], dataloader_idx: int
    ) -> dict[str, Tensor]:
        """Apply batch augmentations to the batch after it is transferred to the device.

        Args:
            batch: A batch of data that needs to be altered or augmented.
            dataloader_idx: The index of the dataloader to which the batch belongs.

        Returns:
            A batch of data.
        """
        if self.trainer:
            if not self.trainer.predicting:
                # Evaluate against flood mask, not water mask
                batch['mask'] = (batch['mask'][:, 1] > 0).long()

        return super().on_after_batch_transfer(batch, dataloader_idx)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""National Agriculture Imagery Program (NAIP) datamodule."""

from typing import Any

import kornia.augmentation as K
from matplotlib.figure import Figure

from ..datasets import (
    NAIP,
    BoundingBox,
    ChesapeakeDC,
    ChesapeakeDE,
    ChesapeakeMD,
    ChesapeakeNY,
    ChesapeakePA,
    ChesapeakeVA,
    ChesapeakeWV,
)
from ..samplers import GridGeoSampler, RandomBatchGeoSampler
from .geo import GeoDataModule


class NAIPChesapeakeDataModule(GeoDataModule):
    """LightningDataModule implementation for the NAIP and Chesapeake datasets.

    Uses the train/val/test splits from the dataset.
    """

    def __init__(
        self,
        batch_size: int = 64,
        patch_size: int | tuple[int, int] = 256,
        length: int | None = None,
        num_workers: int = 0,
        **kwargs: Any,
    ) -> None:
        """Initialize a new NAIPChesapeakeDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            patch_size: Size of each patch, either ``size`` or ``(height, width)``.
            length: Length of each training epoch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.NAIP` (prefix keys with ``naip_``) and
                :class:`~torchgeo.datasets.Chesapeake`
                (prefix keys with ``chesapeake_``).
        """
        self.naip_kwargs = {}
        self.chesapeake_kwargs = {}
        for key, val in kwargs.items():
            if key.startswith('naip_'):
                self.naip_kwargs[key[5:]] = val
            elif key.startswith('chesapeake_'):
                self.chesapeake_kwargs[key[11:]] = val

        super().__init__(
            NAIP, batch_size, patch_size, length, num_workers, **self.naip_kwargs
        )

        self.aug = K.AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std), data_keys=None, keepdim=True
        )

    def setup(self, stage: str) -> None:
        """Set up datasets and samplers.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.naip = NAIP(**self.naip_kwargs)
        dc = ChesapeakeDC(**self.chesapeake_kwargs)
        de = ChesapeakeDE(**self.chesapeake_kwargs)
        md = ChesapeakeMD(**self.chesapeake_kwargs)
        ny = ChesapeakeNY(**self.chesapeake_kwargs)
        pa = ChesapeakePA(**self.chesapeake_kwargs)
        va = ChesapeakeVA(**self.chesapeake_kwargs)
        wv = ChesapeakeWV(**self.chesapeake_kwargs)
        self.chesapeake = dc | de | md | ny | pa | va | wv
        self.dataset = self.naip & self.chesapeake

        roi = self.dataset.bounds
        midx = roi.minx + (roi.maxx - roi.minx) / 2
        midy = roi.miny + (roi.maxy - roi.miny) / 2

        if stage in ['fit']:
            train_roi = BoundingBox(
                roi.minx, midx, roi.miny, roi.maxy, roi.mint, roi.maxt
            )
            self.train_batch_sampler = RandomBatchGeoSampler(
                self.dataset, self.patch_size, self.batch_size, self.length, train_roi
            )
        if stage in ['fit', 'validate']:
            val_roi = BoundingBox(midx, roi.maxx, roi.miny, midy, roi.mint, roi.maxt)
            self.val_sampler = GridGeoSampler(
                self.dataset, self.patch_size, self.patch_size, val_roi
            )
        if stage in ['test']:
            test_roi = BoundingBox(midx, roi.maxx, midy, roi.maxy, roi.mint, roi.maxt)
            self.test_sampler = GridGeoSampler(
                self.dataset, self.patch_size, self.patch_size, test_roi
            )

    def plot(self, *args: Any, **kwargs: Any) -> Figure:
        """Run NAIP plot method.

        Args:
            *args: Arguments passed to plot method.
            **kwargs: Keyword arguments passed to plot method.

        Returns:
            A matplotlib Figure with the image, ground truth, and predictions.

        .. versionadded:: 0.4
        """
        return self.naip.plot(*args, **kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""HySpecNet dataset."""

import os
import re
from collections.abc import Callable, Sequence
from typing import ClassVar

import rasterio as rio
import torch
from einops import rearrange
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from torch import Tensor

from .enmap import EnMAP
from .errors import DatasetNotFoundError, RGBBandsMissingError
from .geo import NonGeoDataset
from .utils import (
    Path,
    disambiguate_timestamp,
    download_url,
    extract_archive,
    percentile_normalization,
)


class HySpecNet11k(NonGeoDataset):
    """HySpecNet-11k dataset.

    `HySpecNet-11k <https://doi.org/10.5061/dryad.fttdz08zh>`__ is a large-scale
    benchmark dataset for hyperspectral image compression and self-supervised learning.
    It is made up of 11,483 nonoverlapping image patches acquired by the
    `EnMAP satellite <https://www.enmap.org/>`_. Each patch is a portion of 128 x 128
    pixels with 224 spectral bands and with a ground sample distance of 30 m.

    To construct HySpecNet-11k, a total of 250 EnMAP tiles acquired during the routine
    operation phase between 2 November 2022 and 9 November 2022 were considered. The
    considered tiles are associated with less than 10% cloud and snow cover. The tiles
    were radiometrically, geometrically and atmospherically corrected (L2A water & land
    product). Then, the tiles were divided into nonoverlapping image patches. The
    cropped patches at the borders of the tiles were eliminated. As a result, more than
    45 patches per tile are obtained, resulting in 11,483 patches for the full dataset.

    We provide predefined splits obtained by randomly dividing HySpecNet into:

    #. a training set that includes 70% of the patches,
    #. a validation set that includes 20% of the patches, and
    #. a test set that includes 10% of the patches.

    Depending on the way that we used for splitting the dataset, we define two
    different splits:

    #. an easy split, where patches from the same tile can be present in different sets
       (patchwise splitting); and
    #. a hard split, where all patches from one tile belong to the same set
       (tilewise splitting).

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2306.00385

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/torchgeo/hyspecnet/resolve/13e110422a6925cbac0f11edff610219b9399227/'
    md5s: ClassVar[dict[str, str]] = {
        'hyspecnet-11k-01.tar.gz': '974aae9197006727b42ec81796049efe',
        'hyspecnet-11k-02.tar.gz': 'f80574485f835b8a263b6c64076c0c62',
        'hyspecnet-11k-03.tar.gz': '6bc1de573f97fa4a75b79719b9270cb3',
        'hyspecnet-11k-04.tar.gz': '2463dc10653cb8be10d44951307c5e7d',
        'hyspecnet-11k-05.tar.gz': '16c1bd9e684673e741c0849bd015c988',
        'hyspecnet-11k-06.tar.gz': '8eef16b67d71af6eb4bc836d294fe3c4',
        'hyspecnet-11k-07.tar.gz': 'f61f0e7d6b05c861e69026b09130a5d6',
        'hyspecnet-11k-08.tar.gz': '19d390bc9e61b85e7d765f3077984976',
        'hyspecnet-11k-09.tar.gz': '197ff47befe5b9de88be5e1321c5ce5d',
        'hyspecnet-11k-10.tar.gz': '9e674cca126a9d139d6584be148d4bac',
        'hyspecnet-11k-splits.tar.gz': '94fad9e3c979c612c29a045406247d6c',
    }

    all_bands = EnMAP.all_bands
    default_bands = EnMAP.default_bands
    rgb_bands = EnMAP.rgb_bands

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        strategy: str = 'easy',
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new HySpecNet11k instance.

        Args:
            root: Root directory where dataset can be found.
            split: One of 'train', 'val', or 'test'.
            strategy: Either 'easy' for patchwise splitting or 'hard' for tilewise
                splitting.
            bands: Bands to return.
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.root = root
        self.split = split
        self.strategy = strategy
        self.bands = bands or self.default_bands
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self.wavelengths = torch.tensor([EnMAP.wavelengths[b] for b in self.bands])
        self.band_indices = [self.all_bands.index(b) + 1 for b in self.bands]

        self._verify()

        path = os.path.join(root, 'hyspecnet-11k', 'splits', strategy, f'{split}.csv')
        with open(path) as f:
            self.files = f.read().strip().split('\n')

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            Length of the dataset.
        """
        return len(self.files)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and label at that index.
        """
        path = self.files[index].replace('DATA.npy', 'SPECTRAL_IMAGE.TIF')
        file = os.path.basename(path)
        match = re.match(EnMAP.filename_regex, file, re.VERBOSE)
        assert match
        mint, maxt = disambiguate_timestamp(match.group('date'), EnMAP.date_format)

        with rio.open(os.path.join(self.root, 'hyspecnet-11k', 'patches', file)) as src:
            minx, maxx = src.bounds.left, src.bounds.right
            miny, maxy = src.bounds.bottom, src.bounds.top
            sample = {
                'image': torch.tensor(src.read(self.band_indices).astype('float32')),
                'x': torch.tensor((minx + maxx) / 2),
                'y': torch.tensor((miny + maxy) / 2),
                't': torch.tensor((mint + maxt) / 2),
                'wavelength': self.wavelengths,
                'res': torch.tensor(30),
            }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        exists = []
        for directory in ['patches', 'splits']:
            path = os.path.join(self.root, 'hyspecnet-11k', directory)
            exists.append(os.path.isdir(path))

        if all(exists):
            return

        for file, md5 in self.md5s.items():
            # Check if the file has already been downloaded
            path = os.path.join(self.root, file)
            if os.path.isfile(path):
                extract_archive(path)
                continue

            # Check if the user requested to download the dataset
            if self.download:
                url = self.url + file
                download_url(url, self.root, md5=md5 if self.checksum else None)
                extract_archive(path)
                continue

            raise DatasetNotFoundError(self)

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: A sample returned by :meth:`__getitem__`.
            suptitle: optional string to use as a suptitle

        Returns:
            A matplotlib Figure with the rendered sample.

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        image = sample['image'][rgb_indices].cpu().numpy()
        image = rearrange(image, 'c h w -> h w c')
        image = percentile_normalization(image)

        fig, ax = plt.subplots()
        ax.imshow(image)
        ax.axis('off')

        if suptitle:
            fig.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench Cloud-S2 dataset."""

import os

from matplotlib.colors import ListedColormap
from torch import Tensor

from .base import CopernicusBenchBase


class CopernicusBenchCloudS2(CopernicusBenchBase):
    """Copernicus-Bench Cloud-S2 dataset.

    Cloud-S2 is a multi-class cloud segmentation dataset derived from
    `CloudSEN12+ <https://www.sciencedirect.com/science/article/pii/S2352340924008163>`_,
    one of the largest Sentinel-2 cloud and cloud shadow detection datasets with
    expert-labeled pixels. We take 25% samples with high-quality labels, and split them
    into 1699/567/551 train/val/test subsets.

    .. list-table:: Classes
       :header-rows: 1

       * - Code
         - Class
         - Description
       * - 0
         - Clear
         - Pixels without cloud and cloud shadow contamination.
       * - 1
         - Thick Cloud
         - Opaque clouds that block all the reflected light from the Earth's surface.
       * - 2
         - Thin Cloud
         - Semitransparent clouds that alter the surface spectral signal but still allow
           to recognize the background. This is the hardest class to identify.
       * - 3
         - Cloud Shadow
         - Dark pixels where light is occluded by thick or thin clouds.

    If you use this dataset in your research, please cite the following papers:

    * https://arxiv.org/abs/2503.11849
    * https://doi.org/10.1016/j.dib.2024.110852

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/wangyi111/Copernicus-Bench/resolve/9d252acd3aa0e3da3128e05c6f028647f0e48e5f/l1_cloud_s2/cloud_s2.zip'
    md5 = '39a1f966e76455549a3e6c209ba751c1'
    zipfile = 'cloud_s2.zip'
    directory = 'cloud_s2'
    filename_regex = r'ROI_\d{5}__(?P<date>\d{8}T\d{6})'
    all_bands = (
        'B01',
        'B02',
        'B03',
        'B04',
        'B05',
        'B06',
        'B07',
        'B08',
        'B8A',
        'B09',
        'B10',
        'B11',
        'B12',
    )
    rgb_bands = ('B04', 'B03', 'B02')
    cmap = ListedColormap(['white', 'yellow', 'green', 'red'])
    classes = ('Clear', 'Thick Cloud', 'Thin Cloud', 'Cloud Shadow')

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        file = self.files[index] + '.tif'
        image_path = os.path.join(self.root, self.directory, 's2_toa', file)
        mask_path = os.path.join(self.root, self.directory, 'cloud', file)
        sample = self._load_image(image_path) | self._load_mask(mask_path)

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""PRISMA datasets."""

from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from .geo import RasterDataset
from .utils import percentile_normalization


class PRISMA(RasterDataset):
    """PRISMA dataset.

    Hyperspectral Precursor and Application Mission
    `PRISMA <https://www.eoportal.org/satellite-missions/prisma-hyperspectral>`__
    (PRecursore IperSpettrale della Missione Applicativa) is a medium-resolution
    hyperspectral imaging satellite, developed, owned, and operated by the
    Italian Space Agency `ASI <https://www.asi.it/en/earth-science/prisma/>`_
    (Agenzia Spaziale Italiana). It is the successor to the discontinued HypSEO
    (Hyperspectral Satellite for Earth Observation) mission.

    PRISMA carries two sensor instruments, the HYC (Hyperspectral Camera) module and the
    PAN (Panchromatic Camera) module. The HYC sensor is a prism spectrometer for two
    bands, VIS/NIR (Visible/Near Infrared) and NIR/SWIR (Near Infrared/Shortwave
    Infrared), with a total of 237 channels across both bands. Its primary mission
    objective is the high resolution hyperspectral imaging of land, vegetation,
    inner waters and coastal zones. The second sensor module, PAN, is a high resolution
    optical imager, and is co-registered with HYC data to allow testing of image fusion
    techniques.

    The HYC module has a spatial resolution of 30 m and operates in two bands, a 66
    channel VIS/NIR band with a spectral interval of 400-1010 nm, and a 171 channel
    NIR/SWIR band with a spectral interval of 920-2505 nm. It uses a pushbroom scanning
    technique with a swath width of 30 km, and a field of regard of 1000 km either side.
    The PAN module also uses a pushbroom scanning technique, with identical swath width
    and field of regard but spatial resolution of 5 m.

    PRISMA is in a sun-synchronous orbit, with an altitude of 614 km, an inclination
    of 98.19° and its LTDN (Local Time on Descending Node) is at 1030 hours.

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1109/IGARSS.2018.8517785

    .. note::
       PRISMA imagery is distributed as HDF5 files. However, TorchGeo does not yet have
       support for reprojecting and windowed reading of HDF5 files. This data loader
       requires you to first convert all files from HDF5 to GeoTIFF using something like
       `this script
       <https://gist.github.com/adamjstewart/7f5324a6a339c20e778f39536402fb4a>`__.

    .. versionadded:: 0.6
    """

    # https://prisma.asi.it/missionselect/docs/PRISMA%20Product%20Specifications_Is2_3.pdf
    #
    # See sections:
    #
    # * 6.3.6: L0A Product Naming Convention
    # * 7.5:   L1 Product Naming Convention
    # * 7.8.5: FKDP, GKDP, ICU-KDP and CDP Products Naming Convention
    filename_glob = 'PRS_*'
    filename_regex = r"""
        ^PRS
        _(?P<level>[A-Z\d]+)
        _(?P<product>[A-Z]+)
        (_(?P<order>[A-Z_]+))?
        _(?P<start>\d{14})
        _(?P<stop>\d{14})
        _(?P<version>\d{4})
        (_(?P<valid>\d))?
        \.
    """
    date_format = '%Y%m%d%H%M%S'

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        # RGB band indices based on https://doi.org/10.3390/rs14164080
        rgb_indices = [34, 23, 11]
        image = sample['image'][rgb_indices].permute(1, 2, 0).float()
        image = percentile_normalization(image, axis=(0, 1))

        fig, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')

        if show_titles:
            ax.set_title('Image')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Trainers for I/O benchmarking."""

from typing import Any

import lightning
import torch
from torch import Tensor
from torch.optim import SGD

from .base import BaseTask


class IOBenchTask(BaseTask):
    """I/O benchmarking.

    .. versionadded:: 0.6
    """

    def configure_models(self) -> None:
        """No-op."""

    def configure_optimizers(
        self,
    ) -> 'lightning.pytorch.utilities.types.OptimizerLRScheduler':
        """Initialize the optimizer.

        Returns:
            Optimizer.
        """
        optimizer = SGD([torch.tensor(0.0, requires_grad=True)], lr=0)
        return {'optimizer': optimizer}

    def training_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """No-op.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Zero.
        """
        return torch.tensor(0.0, requires_grad=True)

    def validation_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """No-op.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """No-op.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """No-op.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 4 dataset."""

import os
from typing import ClassVar

from .base import SpaceNet


class SpaceNet4(SpaceNet):
    """SpaceNet 4: Off-Nadir Buildings Dataset.

    `SpaceNet 4 <https://spacenet.ai/off-nadir-building-detection/>`_ is a
    dataset of 27 WV-2 imagery captured at varying off-nadir angles and
    associated building footprints over the city of Atlanta. The off-nadir angle
    ranges from 7 degrees to 54 degrees.

    Dataset features:

    * No. of chipped images: 28,728 (PAN/MS/PS-RGBNIR)
    * No. of label files: 1064
    * No. of building footprints: >120,000
    * Area Coverage: 665 sq km
    * Chip size: 225 x 225 (MS), 900 x 900 (PAN/PS-RGBNIR)

    Dataset format:

    * Imagery - Worldview-2 GeoTIFFs

        * PAN.tif (Panchromatic)
        * MS.tif (Multispectral)
        * PS-RGBNIR (Pansharpened RGBNIR)

    * Labels - GeoJSON

        * labels.geojson

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/1903.12239
    """

    directory_glob = os.path.join('**', '{product}')
    file_regex = r'_(\d+_\d+)\.'
    dataset_id = 'SN4_buildings'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            6: [
                'Atlanta_nadir7_catid_1030010003D22F00.tar.gz',
                'Atlanta_nadir8_catid_10300100023BC100.tar.gz',
                'Atlanta_nadir10_catid_1030010003993E00.tar.gz',
                'Atlanta_nadir10_catid_1030010003CAF100.tar.gz',
                'Atlanta_nadir13_catid_1030010002B7D800.tar.gz',
                'Atlanta_nadir14_catid_10300100039AB000.tar.gz',
                'Atlanta_nadir16_catid_1030010002649200.tar.gz',
                'Atlanta_nadir19_catid_1030010003C92000.tar.gz',
                'Atlanta_nadir21_catid_1030010003127500.tar.gz',
                'Atlanta_nadir23_catid_103001000352C200.tar.gz',
                'Atlanta_nadir25_catid_103001000307D800.tar.gz',
                'Atlanta_nadir27_catid_1030010003472200.tar.gz',
                'Atlanta_nadir29_catid_1030010003315300.tar.gz',
                'Atlanta_nadir30_catid_10300100036D5200.tar.gz',
                'Atlanta_nadir32_catid_103001000392F600.tar.gz',
                'Atlanta_nadir34_catid_1030010003697400.tar.gz',
                'Atlanta_nadir36_catid_1030010003895500.tar.gz',
                'Atlanta_nadir39_catid_1030010003832800.tar.gz',
                'Atlanta_nadir42_catid_10300100035D1B00.tar.gz',
                'Atlanta_nadir44_catid_1030010003CCD700.tar.gz',
                'Atlanta_nadir46_catid_1030010003713C00.tar.gz',
                'Atlanta_nadir47_catid_10300100033C5200.tar.gz',
                'Atlanta_nadir49_catid_1030010003492700.tar.gz',
                'Atlanta_nadir50_catid_10300100039E6200.tar.gz',
                'Atlanta_nadir52_catid_1030010003BDDC00.tar.gz',
                'Atlanta_nadir53_catid_1030010003193D00.tar.gz',
                'Atlanta_nadir53_catid_1030010003CD4300.tar.gz',
                'geojson.tar.gz',
            ]
        },
        'test': {6: ['SN4_buildings_AOI_6_Atlanta_test_public.tar.gz']},
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            6: [
                'd41ab6ec087b07e1e046c55d1fa5754b',
                '72f04a7c0c34dd4595c181ee1ae6cb4c',
                '89559f42ac11a8de570cef9802a577ad',
                '5489ac756249c336ea506ef0acb3c09d',
                'bd9ed231cedd8631683ea51ea0602de1',
                'c497a8a448ed7ccdf63e7706507c0603',
                '45d54eeecefdc60aa38320be6f29a17c',
                '611528c0188bbc7e9cdf98609c6b0c49',
                '532fbf1ca73d3d2e8b03c585f61b7316',
                '538f48429b0968b6cfad97eb61fa8de1',
                '3c48e94bc6d9e66e27c3a9bc8d35d65d',
                'b78cdf951e7bf4fedbe9259abd1e047a',
                'f307ce3c623d12d5a2fd5acb1e0607e0',
                '9a17574332cd5513d68a0bcc9c607bdd',
                'fe905ca809f7bd2ceef75bde23c326f3',
                'd9f2e4a5c8462f6f9f7d5c573d9a1dc6',
                'f9425ff38dc82bf0e8f25a6287ff1ad1',
                '7a6005d6fd972d5ce04caf9b42b36897',
                '7c5aa16bb64cacf766cf88f89b3093bd',
                '8f7e959eb0156ad2dfb0b966a1de06a9',
                '62c4babcbe70034b7deb7c14d5ff61c2',
                '8001d75f67534edf6932242324b8c1a7',
                'bc299cb5de432b5f5a1ce65a3bdb0abc',
                'd7640eda7c4efaf825665e853037bec9',
                'd4e1931551e9d3c6fd9bf1d8adfd07a0',
                'b313e23ead8fe6e2c8671a49f2c9de37',
                '3bd8f07ad57bff841d0cf91c91c6f5ed',
                '2556339e26a09e57559452eb240ef29c',
            ]
        },
        'test': {6: ['0ec3874bfc19aed63b33ac47b039aace']},
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {'train': [6], 'test': [6]}
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['MS', 'PAN', 'Pan-Sharpen'],
        'test': ['MS', 'PAN', 'Pan-Sharpen'],
    }
    valid_masks = (os.path.join('geojson', 'spacenet-buildings'),)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# Code based on https://github.com/antofuller/CROMA under MIT License

"""CROMA model."""

import itertools
import math
from collections.abc import Sequence
from typing import Any

import torch
from einops import rearrange
from torch import Tensor, einsum, nn
from torchvision.models._api import Weights, WeightsEnum


class CROMA(nn.Module):
    """Pretrained CROMA model.

    Corresponds to the pretrained CROMA model found in the CROMA repository:

    * https://github.com/antofuller/CROMA/blob/main/pretrain_croma.py

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2311.00566
    """

    valid_modalities = ('sar', 'optical')

    def __init__(
        self,
        modalities: Sequence[str] = ['sar', 'optical'],
        encoder_dim: int = 768,
        encoder_depth: int = 12,
        num_heads: int = 16,
        patch_size: int = 8,
        image_size: int = 120,
    ) -> None:
        """Initialize the CROMA model.

        Args:
            modalities: List of modalities used during forward pass, list can contain
                'sar', 'optical', or both.
            encoder_dim: Dimension of the encoder.
            encoder_depth: Depth of the encoder.
            num_heads: Number of heads for the multi-head attention, should be power of 2.
            patch_size: Size of the patches.
            image_size: Size of the input images, CROMA was trained on 120x120 images,
                must be a multiple of 8.

        Raises:
            AssertionError: If any arguments are not valid.
        """
        super().__init__()
        for modality in modalities:
            assert modality in self.valid_modalities, (
                f'{modality} is not a valid modality'
            )

        assert image_size % 8 == 0, 'image_size must be a multiple of 8'
        assert num_heads % 2 == 0, 'num_heads must be a power of 2'

        self.modalities = modalities
        self.encoder_dim = encoder_dim
        self.encoder_depth = encoder_depth
        self.num_heads = num_heads
        self.patch_size = patch_size
        self.image_size = image_size

        self.num_patches = int((image_size / 8) ** 2)
        self.s1_channels = 2  # fixed at 2 SAR backscatter channels
        self.s2_channels = 12  # fixed at 12 multispectral optical channels

        self.attn_bias = get_2dalibi(
            num_heads=self.num_heads, num_patches=self.num_patches
        )

        def initialize_encoder(
            encoder_dim: int, encoder_depth: int, in_channels: int
        ) -> tuple[nn.Module, nn.Module]:
            """Initialize the encoder and GAP-FFN for a given modality.

            Args:
                encoder_dim: Dimension of the encoder.
                encoder_depth: Depth of the encoder.
                in_channels: Number of input channels.

            Returns:
                Tuple containing the encoder and GAP-FFN.
            """
            encoder = ViT(dim=encoder_dim, depth=encoder_depth, in_channels=in_channels)
            gap_ffn = nn.Sequential(
                nn.LayerNorm(encoder_dim),
                nn.Linear(encoder_dim, int(4 * encoder_dim)),
                nn.GELU(),
                nn.Linear(int(4 * encoder_dim), encoder_dim),
            )
            return encoder, gap_ffn

        if 'sar' in modalities:
            self.s1_encoder, self.s1_GAP_FFN = initialize_encoder(
                encoder_dim, int(encoder_depth / 2), self.s1_channels
            )
        if 'optical' in modalities:
            self.s2_encoder, self.s2_GAP_FFN = initialize_encoder(
                encoder_dim, encoder_depth, self.s2_channels
            )
        if set(self.modalities) == {'sar', 'optical'}:
            self.joint_encoder = BaseTransformerCrossAttn(
                dim=encoder_dim, depth=int(encoder_depth / 2), num_heads=num_heads
            )

    def forward(
        self, x_sar: Tensor | None = None, x_optical: Tensor | None = None
    ) -> dict[str, Tensor]:
        """Forward pass of the CROMA model.

        Args:
            x_sar: Input mini-batch of SAR images [B, 2, H, W].
            x_optical: Input mini-batch of optical images [B, 12, H, W].
        """
        return_dict: dict[str, Tensor] = {}

        if 'sar' in self.modalities and x_sar is not None:
            sar_encodings = self.s1_encoder(imgs=x_sar, attn_bias=self.attn_bias)
            sar_GAP = self.s1_GAP_FFN(sar_encodings.mean(dim=1))
            return_dict['sar_encodings'] = sar_encodings
            return_dict['sar_GAP'] = sar_GAP

        if 'optical' in self.modalities and x_optical is not None:
            optical_encodings = self.s2_encoder(
                imgs=x_optical, attn_bias=self.attn_bias
            )
            optical_GAP = self.s2_GAP_FFN(optical_encodings.mean(dim=1))
            return_dict['optical_encodings'] = optical_encodings
            return_dict['optical_GAP'] = optical_GAP

        if set(self.modalities) == {'sar', 'optical'}:
            joint_encodings = self.joint_encoder(
                x=sar_encodings,
                context=optical_encodings,
                relative_position_bias=self.attn_bias,
            )
            joint_GAP = joint_encodings.mean(dim=1)
            return_dict['joint_encodings'] = joint_encodings
            return_dict['joint_GAP'] = joint_GAP

        return return_dict


def get_2dalibi(num_heads: int, num_patches: int) -> Tensor:
    """Get 2D relative position bias for the attention layer.

    Args:
        num_heads: Number of heads for the multi-head attention.
        num_patches: Number of patches.

    Returns:
        2D relative position bias tensor.
    """
    # inspired by: https://github.com/ofirpress/attention_with_linear_biases
    points = list(
        itertools.product(
            range(int(math.sqrt(num_patches))), range(int(math.sqrt(num_patches)))
        )
    )

    def get_slopes(n: int) -> list[float]:
        start = 2 ** (-(2 ** -(math.log2(n) - 3)))
        ratio = start
        return [start * ratio**i for i in range(n)]

    slopes = torch.Tensor(get_slopes(num_heads)).unsqueeze(1)
    idxs = []
    for p1 in points:
        for p2 in points:
            dist = math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)
            idxs.append(dist * slopes * -1)
    all_bias = torch.cat(idxs, dim=1)
    return all_bias.view(1, num_heads, num_patches, num_patches)


class FFN(nn.Module):
    """Feed-forward network for the transformer."""

    def __init__(self, dim: int, mult: int = 4, dropout: float = 0.0) -> None:
        """Initialize the feed-forward network.

        Args:
            dim: Dimension of the input.
            mult: Multiplier for the inner dimension of the feed-forward network.
            dropout: Dropout probability
        """
        super().__init__()
        inner_dim = int(dim * mult)

        self.net = nn.Sequential(
            nn.Linear(dim, inner_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(inner_dim, dim),
        )
        self.input_norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass of the feed-forward network.

        Args:
            x: Input tensor.

        Returns:
            Output tensor.
        """
        x = self.input_norm(x)
        x = self.net(x)
        return x


class Attention(nn.Module):
    """Multi-head attention layer for the transformer."""

    def __init__(self, dim: int, num_heads: int = 8, dropout: float = 0.0) -> None:
        """Initialize the multi-head attention layer.

        Args:
            dim: Dimension of the input.
            num_heads: Number of heads for the multi-head attention.
            dropout: Dropout probability.
        """
        super().__init__()
        self.num_heads = num_heads
        assert dim % num_heads == 0, 'dim must be evenly divisible by num_heads'
        dim_head = int(dim / num_heads)
        self.scale = dim_head**-0.5

        self.to_qkv = nn.Linear(dim, dim * 3, bias=False)
        self.to_out = nn.Linear(dim, dim)
        self.input_norm = nn.LayerNorm(dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: Tensor, relative_position_bias: Tensor) -> Tensor:
        """Forward pass of the multi-head attention layer.

        Args:
            x: Input tensor.
            relative_position_bias: Relative position bias tensor.

        Returns:
            Output tensor.
        """
        x = self.input_norm(x)
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.num_heads), (q, k, v)
        )

        attention_scores = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
        attention_scores = attention_scores + relative_position_bias

        attn = attention_scores.softmax(dim=-1)
        attn = self.dropout(attn)

        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        out = self.to_out(out)
        return out


class CrossAttention(nn.Module):
    """Cross-attention layer for the transformer."""

    def __init__(self, dim: int, num_heads: int = 8, dropout: float = 0.0) -> None:
        """Initialize the cross-attention layer.

        Args:
            dim: Dimension of the input.
            num_heads: Number of heads for the multi-head attention.
            dropout: Dropout probability.

        Raises:
            AssertionError: If the dimension is not evenly divisible by the number of heads.
        """
        super().__init__()
        self.num_heads = num_heads
        assert dim % num_heads == 0, 'dim must be evenly divisible by num_heads'
        dim_head = int(dim / num_heads)
        self.scale = dim_head**-0.5

        self.to_q = nn.Linear(dim, dim, bias=False)
        self.to_k = nn.Linear(dim, dim, bias=False)
        self.to_v = nn.Linear(dim, dim, bias=False)

        self.to_out = nn.Linear(dim, dim)
        self.input_norm = nn.LayerNorm(dim)
        self.dropout = nn.Dropout(dropout)

    def forward(
        self, x: Tensor, context: Tensor, relative_position_bias: Tensor
    ) -> Tensor:
        """Forward pass of the cross-attention layer.

        Args:
            x: Input tensor.
            context: Context tensor.
            relative_position_bias: Relative position bias tensor.

        Returns:
            Output tensor.
        """
        x = self.input_norm(x)
        context = self.input_norm(context)

        q = self.to_q(x)
        k = self.to_k(context)
        v = self.to_v(context)

        q, k, v = map(
            lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.num_heads), (q, k, v)
        )

        attention_scores = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
        attention_scores = attention_scores + relative_position_bias

        attn = attention_scores.softmax(dim=-1)
        attn = self.dropout(attn)

        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        out = self.to_out(out)
        return out


class BaseTransformer(nn.Module):
    """Base transformer model."""

    def __init__(
        self,
        dim: int,
        depth: int,
        num_heads: int = 8,
        attn_dropout: float = 0.0,
        ff_dropout: float = 0.0,
        ff_mult: int = 4,
        final_norm: bool = True,
    ) -> None:
        """Initialize the base transformer model.

        Args:
            dim: Dimension of the input.
            depth: Depth of the transformer.
            num_heads: Number of heads for the multi-head attention.
            attn_dropout: Dropout probability for the attention layer.
            ff_dropout: Dropout probability for the feed-forward network.
            ff_mult: Multiplier for the inner dimension of the feed-forward network.
            final_norm: Whether to apply a final layer normalization.
        """
        super().__init__()
        self.final_norm = final_norm
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(dim=dim, num_heads=num_heads, dropout=attn_dropout),
                        FFN(dim=dim, mult=ff_mult, dropout=ff_dropout),
                    ]
                )
            )

        if self.final_norm:
            self.norm_out = nn.LayerNorm(dim)

    def forward(self, x: Tensor, relative_position_bias: Tensor) -> Tensor:
        """Forward pass of the base transformer model.

        Args:
            x: Input tensor.
            relative_position_bias: whether to use relative position bias.
        """
        for self_attn, ffn in self.layers:
            x = self_attn(x, relative_position_bias) + x
            x = ffn(x) + x

        x = self.norm_out(x) if self.final_norm else x
        return x


class BaseTransformerCrossAttn(nn.Module):
    """Base transformer model with cross-attention."""

    def __init__(
        self,
        dim: int,
        depth: int,
        num_heads: int = 8,
        attn_dropout: float = 0.0,
        ff_dropout: float = 0.0,
        ff_mult: int = 4,
    ) -> None:
        """Initialize the base transformer model with cross-attention.

        Args:
            dim: Dimension of the input.
            depth: Depth of the transformer.
            num_heads: Number of heads for the multi-head attention.
            attn_dropout: Dropout probability for the attention layer.
            ff_dropout: Dropout probability for the feed-forward network.
            ff_mult: Multiplier for the inner dimension of the feed-forward network.
        """
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(dim=dim, num_heads=num_heads, dropout=attn_dropout),
                        CrossAttention(
                            dim=dim, num_heads=num_heads, dropout=attn_dropout
                        ),
                        FFN(dim=dim, mult=ff_mult, dropout=ff_dropout),
                    ]
                )
            )

        self.norm_out = nn.LayerNorm(dim)

    def forward(
        self, x: Tensor, context: Tensor, relative_position_bias: Tensor
    ) -> Tensor:
        """Forward pass of the base transformer model with cross-attention.

        Args:
            x: Input tensor.
            context: Context tensor.
            relative_position_bias: Relative position bias tensor.

        Returns:
            Output tensor.
        """
        for self_attn, cross_attn, ffn in self.layers:
            x = self_attn(x, relative_position_bias) + x
            x = cross_attn(x, context, relative_position_bias) + x
            x = ffn(x) + x

        x = self.norm_out(x)
        return x


class ViT(nn.Module):
    """Vision Transformer model."""

    def __init__(self, dim: int, depth: int, in_channels: int) -> None:
        """Initialize the vision transformer model.

        Args:
            dim: Dimension of the input.
            depth: Depth of the transformer.
            in_channels: Number of input channels.
        """
        super().__init__()
        self.depth = depth
        self.in_channels = in_channels
        self.dim = dim
        self.num_heads = 16  # always 16, for base and large models
        self.patch_size = 8  # always 8, for base and large models

        pixels_per_patch = int(self.patch_size * self.patch_size * in_channels)
        self.linear_input = nn.Linear(pixels_per_patch, self.dim)
        self.transformer = BaseTransformer(
            dim=self.dim, depth=self.depth, num_heads=self.num_heads
        )

    def forward(self, imgs: Tensor, attn_bias: Tensor) -> Tensor:
        """Forward pass of the vision transformer model.

        Args:
            imgs: Input tensor.
            attn_bias: Relative position bias tensor.

        Returns:
            Output tensor.
        """
        x = rearrange(
            imgs,
            'b c (h i) (w j) -> b (h w) (c i j)',
            i=self.patch_size,
            j=self.patch_size,
        )
        # x is shape -> (bsz, num_patches, self.channels*self.patch_size*self.patch_size)

        x = self.linear_input(x)
        x = self.transformer(x, relative_position_bias=attn_bias)
        return x


class CROMABase_Weights(WeightsEnum):  # type: ignore[misc]
    """CROMA base model weights.

    .. versionadded:: 0.7
    """

    CROMA_VIT = Weights(
        url='https://hf.co/torchgeo/croma/resolve/387883f08af79d777167519c57cd826eda89a16f/CROMA_base-0238d814.pt',
        transforms=None,
        meta={
            'dataset': 'SSL4EO',
            'model': 'vit',
            'publication': 'https://arxiv.org/abs/2311.00566',
            'repo': 'https://github.com/antofuller/CROMA',
            'ssl_method': 'croma',
        },
    )


class CROMALarge_Weights(WeightsEnum):  # type: ignore[misc]
    """CROMA large model weights.

    .. versionadded:: 0.7
    """

    CROMA_VIT = Weights(
        url='https://huggingface.co/torchgeo/croma/resolve/92cb1a0f4e34c6c01558baf070197c01255382f6/CROMA_large-921e69ad.pt',
        transforms=None,
        meta={
            'dataset': 'SSL4EO',
            'model': 'vit',
            'publication': 'https://arxiv.org/abs/2311.00566',
            'repo': 'https://github.com/antofuller/CROMA',
            'ssl_method': 'croma',
        },
    )


def load_weights(model: CROMA, weights: WeightsEnum) -> None:
    """Load weights from a WeightsEnum object.

    Args:
        model: Model to load the weights into.
        weights: Weights to load.

    Raises:
        AssertionError: If there are missing or unexpected keys.
    """
    state_dict = weights.get_state_dict(progress=True)
    missing_keys, unexpected_keys = [], []

    if 'sar' in model.modalities:
        miss_key, unexp_key = model.s1_encoder.load_state_dict(
            state_dict['s1_encoder'], strict=False
        )
        missing_keys.extend(miss_key)
        unexpected_keys.extend(unexp_key)
        miss_key, unexp_key = model.s1_GAP_FFN.load_state_dict(
            state_dict['s1_GAP_FFN'], strict=False
        )
        missing_keys.extend(miss_key)
        unexpected_keys.extend(unexp_key)

    if 'optical' in model.modalities:
        miss_key, unexp_key = model.s2_encoder.load_state_dict(
            state_dict['s2_encoder'], strict=False
        )
        missing_keys.extend(miss_key)
        unexpected_keys.extend(unexp_key)
        miss_key, unexp_key = model.s2_GAP_FFN.load_state_dict(
            state_dict['s2_GAP_FFN'], strict=False
        )
        missing_keys.extend(miss_key)
        unexpected_keys.extend(unexp_key)

    if set(model.modalities) == {'sar', 'optical'}:
        miss_key, unexp_key = model.joint_encoder.load_state_dict(
            state_dict['joint_encoder'], strict=False
        )
        missing_keys.extend(miss_key)
        unexpected_keys.extend(unexp_key)

    assert not missing_keys, f'Missing keys: {missing_keys}'
    assert not unexpected_keys, f'Unexpected keys: {unexpected_keys}'


def croma_base(
    weights: CROMABase_Weights | None = None, *args: Any, **kwargs: Any
) -> CROMA:
    """CROMA base model.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2311.00566

    .. versionadded:: 0.7

    Args:
        weights: Pretrained weights to load.
        *args: Additional arguments to pass to :class:CROMA.`
        **kwargs: Additional keyword arguments to pass to :class:CROMA.`

    Returns:
        CROMA base model.
    """
    kwargs |= {
        'encoder_dim': 768,
        'encoder_depth': 12,
        'num_heads': 16,
        'patch_size': 8,
    }
    model = CROMA(*args, **kwargs)
    if weights:
        load_weights(model, weights)
    return model


def croma_large(
    weights: CROMALarge_Weights | None = None, *args: Any, **kwargs: Any
) -> CROMA:
    """CROMA large model.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2311.00566

    .. versionadded:: 0.7

    Args:
        weights: Pretrained weights to load.
        *args: Additional arguments to pass to :class:CROMA.`
        **kwargs: Additional keyword arguments to pass to :class:CROMA.`

    Returns:
        CROMA large model.
    """
    kwargs |= {
        'encoder_dim': 1024,
        'encoder_depth': 24,
        'num_heads': 16,
        'patch_size': 8,
    }
    model = CROMA(*args, **kwargs)
    if weights:
        load_weights(model, weights)
    return model



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""GeoNRW datamodule."""

import os
from typing import Any

import kornia.augmentation as K
from torch.utils.data import Subset

from ..datasets import GeoNRW
from .geo import NonGeoDataModule
from .utils import group_shuffle_split


class GeoNRWDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the GeoNRW dataset.

    Implements 80/20 train/val splits based on city locations.
    See :func:`setup` for more details.

    .. versionadded:: 0.6
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, size: int = 256, **kwargs: Any
    ) -> None:
        """Initialize a new GeoNRWDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            size: resize images of input size 1000x1000 to size x size
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.GeoNRW`.
        """
        super().__init__(GeoNRW, batch_size, num_workers, **kwargs)

        self.train_aug = K.AugmentationSequential(
            K.Resize(size),
            K.RandomHorizontalFlip(p=0.5),
            K.RandomVerticalFlip(p=0.5),
            data_keys=None,
            keepdim=True,
        )

        self.aug = K.AugmentationSequential(
            K.Resize(size), data_keys=None, keepdim=True
        )

        self.size = size

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit', 'validate']:
            dataset = GeoNRW(split='train', **self.kwargs)
            city_paths = [os.path.dirname(path) for path in dataset.file_list]
            train_indices, val_indices = group_shuffle_split(
                city_paths, test_size=0.2, random_state=0
            )
            self.train_dataset = Subset(dataset, train_indices)
            self.val_dataset = Subset(dataset, val_indices)
        if stage in ['test']:
            self.test_dataset = GeoNRW(split='test', **self.kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""GeoNRW dataset."""

import os
from collections.abc import Callable
from glob import glob
from typing import ClassVar

import matplotlib
import matplotlib.cm
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor
from torchvision import transforms

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_and_extract_archive, extract_archive


class GeoNRW(NonGeoDataset):
    """GeoNRW dataset.

    This datasets contains RGB, DEM and segmentation label data from North Rhine-Westphalia, Germany.

    Dataset features:

    * 7298 training and 485 test samples
    * RGB images, 1000x1000px normalized to [0, 1]
    * DEM images, unnormalized
    * segmentation labels

    Dataset format:

    * RGB images are three-channel jp2
    * DEM images are single-channel tif
    * segmentation labels are single-channel tif

    Dataset classes:

    0. background
    1. forest
    2. water
    3. agricultural
    4. residential,commercial,industrial
    5. grassland,swamp,shrubbery
    6. railway,trainstation
    7. highway,squares
    8. airport,shipyard
    9. roads
    10. buildings

    Additional information about the dataset can be found `on this site <https://ieee-dataport.org/open-access/geonrw>`__.

    If you use this dataset in your research, please cite the following paper:

    * https://ieeexplore.ieee.org/document/9406194


    .. versionadded:: 0.6
    """

    # Splits taken from https://github.com/gbaier/geonrw/blob/ecfcdbca8cfaaeb490a9c6916980f385b9f3941a/pytorch/nrw.py#L48

    splits = ('train', 'test')

    train_list: tuple[str, ...] = (
        'aachen',
        'bergisch',
        'bielefeld',
        'bochum',
        'bonn',
        'borken',
        'bottrop',
        'coesfeld',
        'dortmund',
        'dueren',
        'duisburg',
        'ennepetal',
        'erftstadt',
        'essen',
        'euskirchen',
        'gelsenkirchen',
        'guetersloh',
        'hagen',
        'hamm',
        'heinsberg',
        'herford',
        'hoexter',
        'kleve',
        'koeln',
        'krefeld',
        'leverkusen',
        'lippetal',
        'lippstadt',
        'lotte',
        'moenchengladbach',
        'moers',
        'muelheim',
        'muenster',
        'oberhausen',
        'paderborn',
        'recklinghausen',
        'remscheid',
        'siegen',
        'solingen',
        'wuppertal',
    )

    test_list: tuple[str, ...] = ('duesseldorf', 'herne', 'neuss')

    classes = (
        'background',
        'forest',
        'water',
        'agricultural',
        'residential,commercial,industrial',
        'grassland,swamp,shrubbery',
        'railway,trainstation',
        'highway,squares',
        'airport,shipyard',
        'roads',
        'buildings',
    )

    colormap = mcolors.ListedColormap(
        [
            '#000000',  # matplotlib black for background
            '#2ca02c',  # matplotlib green for forest
            '#1f77b4',  # matplotlib blue for water
            '#8c564b',  # matplotlib brown for agricultural
            '#7f7f7f',  # matplotlib gray residential_commercial_industrial
            '#bcbd22',  # matplotlib olive for grassland_swamp_shrubbery
            '#ff7f0e',  # matplotlib orange for railway_trainstation
            '#9467bd',  # matplotlib purple for highway_squares
            '#17becf',  # matplotlib cyan for airport_shipyard
            '#d62728',  # matplotlib red for roads
            '#e377c2',  # matplotlib pink for buildings
        ]
    )

    readers: ClassVar[dict[str, Callable[[str], Image.Image]]] = {
        'rgb': lambda path: Image.open(path).convert('RGB'),
        'dem': lambda path: Image.open(path).copy(),
        'seg': lambda path: Image.open(path).convert('I;16'),
    }

    modality_filenames: ClassVar[dict[str, Callable[[list[str]], str]]] = {
        'rgb': lambda utm_coords: '{}_{}_rgb.jp2'.format(*utm_coords),
        'dem': lambda utm_coords: '{}_{}_dem.tif'.format(*utm_coords),
        'seg': lambda utm_coords: '{}_{}_seg.tif'.format(*utm_coords),
    }

    modalities: tuple[str, ...] = ('rgb', 'dem', 'seg')

    url = 'https://hf.co/datasets/torchgeo/geonrw/resolve/3cb6bdf2a615b9e526c7dcff85fd1f20728081b7/{}'

    filename = 'nrw_dataset.tar.gz'
    md5 = 'd56ab50098d5452c33d08ff4e99ce281'

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize the GeoNRW dataset.

        Args:
            root: root directory where dataset can be found
            split: one of "train", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` argument is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits, f'split must be one of {self.splits}'

        self.root = root
        self.split = split
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self.city_names = self.test_list if split == 'test' else self.train_list

        self._verify()

        self.file_list = self._get_file_list()

    def _get_file_list(self) -> list[str]:
        """Get a list of files for cities in the dataset split.

        Returns:
            list of filenames in the dataset split
        """
        file_list: list[str] = []
        for cn in self.city_names:
            pattern = os.path.join(self.root, cn, '*rgb.jp2')
            file_list.extend(glob(pattern))
        return sorted(file_list)

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.file_list)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        to_tensor = transforms.ToTensor()

        path: str = self.file_list[index]
        utm_coords = os.path.basename(path).split('_')[:2]
        base_dir = os.path.dirname(path)

        sample: dict[str, Tensor] = {}
        for modality in self.modalities:
            modality_path = os.path.join(
                base_dir, self.modality_filenames[modality](utm_coords)
            )
            sample[modality] = to_tensor(self.readers[modality](modality_path))

        # rename to torchgeo standard keys
        sample['image'] = sample.pop('rgb').float()
        sample['mask'] = sample.pop('seg').long().squeeze(0)

        if self.transforms:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # check if city names directories exist
        all_exist = all(
            os.path.exists(os.path.join(self.root, cn)) for cn in self.city_names
        )
        if all_exist:
            return

        # Check if the tar file has been downloaded
        if os.path.exists(os.path.join(self.root, self.filename)):
            extract_archive(os.path.join(self.root, self.filename), self.root)
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        download_and_extract_archive(
            self.url.format(self.filename),
            download_root=self.root,
            md5=self.md5 if self.checksum else None,
        )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        showing_predictions = 'prediction' in sample
        ncols = 3
        if showing_predictions:
            prediction = sample['prediction'].long()
            ncols += 1

        fig, axs = plt.subplots(
            nrows=1, ncols=ncols, figsize=(ncols * 5, 10), sharex=True
        )

        axs[0].imshow(sample['image'].permute(1, 2, 0))
        axs[0].axis('off')
        axs[1].imshow(sample['dem'].squeeze(0), cmap='gray')
        axs[1].axis('off')
        axs[2].imshow(
            sample['mask'].squeeze(0),
            self.colormap,
            vmin=0,
            vmax=10,
            interpolation='none',
        )
        axs[2].axis('off')

        if showing_predictions:
            axs[3].imshow(
                prediction.squeeze(0),
                self.colormap,
                vmin=0,
                vmax=10,
                interpolation='none',
            )

        # show classes in legend
        if show_titles:
            patches = [matplotlib.patches.Patch(color=c) for c in self.colormap.colors]  # type: ignore
            axs[2].legend(
                patches, self.classes, loc='center left', bbox_to_anchor=(1, 0.5)
            )

        if show_titles:
            axs[0].set_title('RGB Image')
            axs[1].set_title('DEM')
            axs[2].set_title('Labels')

        if suptitle is not None:
            fig.suptitle(suptitle, y=0.8)

        fig.tight_layout()

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TorchGeo indices transforms.

For more information about indices see the following references:
- https://www.indexdatabase.de/db/i.php
- https://github.com/awesome-spectral-indices/awesome-spectral-indices
"""

import torch
from kornia.augmentation import IntensityAugmentationBase2D
from torch import Tensor

_EPSILON = 1e-10


class AppendNormalizedDifferenceIndex(IntensityAugmentationBase2D):
    r"""Append normalized difference index as channel to image tensor.

    Computes the following index:

    .. math::

       \text{NDI} = \frac{A - B}{A + B}

    .. versionadded:: 0.2
    """

    def __init__(self, index_a: int, index_b: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_a: reference band channel index
            index_b: difference band channel index
        """
        super().__init__(p=1)
        self.flags = {'index_a': index_a, 'index_b': index_b}

    def apply_transform(
        self,
        input: Tensor,
        params: dict[str, Tensor],
        flags: dict[str, int],
        transform: Tensor | None = None,
    ) -> Tensor:
        """Apply the transform.

        Args:
            input: the input tensor
            params: generated parameters
            flags: static parameters
            transform: the geometric transformation tensor

        Returns:
            the augmented input
        """
        band_a = input[..., flags['index_a'], :, :]
        band_b = input[..., flags['index_b'], :, :]
        ndi = (band_a - band_b) / (band_a + band_b + _EPSILON)
        ndi = torch.unsqueeze(ndi, -3)
        input = torch.cat((input, ndi), dim=-3)
        return input


class AppendNBR(AppendNormalizedDifferenceIndex):
    r"""Normalized Burn Ratio (NBR).

    Computes the following index:

    .. math::

       \text{NBR} = \frac{\text{NIR} - \text{SWIR}}{\text{NIR} + \text{SWIR}}

    If you use this index in your research, please cite the following paper:

    * https://www.yumpu.com/en/document/view/24226870/the-normalized-burn-ratio-and-relationships-to-burn-severity-/7

    .. versionadded:: 0.2
    """

    def __init__(self, index_nir: int, index_swir: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the Near Infrared (NIR) band in the image
            index_swir: index of the Short-wave Infrared (SWIR) band in the image
        """
        super().__init__(index_a=index_nir, index_b=index_swir)


class AppendNDBI(AppendNormalizedDifferenceIndex):
    r"""Normalized Difference Built-up Index (NDBI).

    Computes the following index:

    .. math::

       \text{NDBI} = \frac{\text{SWIR} - \text{NIR}}{\text{SWIR} + \text{NIR}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1080/01431160304987
    """

    def __init__(self, index_swir: int, index_nir: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_swir: index of the Short-wave Infrared (SWIR) band in the image
            index_nir: index of the Near Infrared (NIR) band in the image
        """
        super().__init__(index_a=index_swir, index_b=index_nir)


class AppendNDSI(AppendNormalizedDifferenceIndex):
    r"""Normalized Difference Snow Index (NDSI).

    Computes the following index:

    .. math::

       \text{NDSI} = \frac{\text{G} - \text{SWIR}}{\text{G} + \text{SWIR}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1109/IGARSS.1994.399618
    """

    def __init__(self, index_green: int, index_swir: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_green: index of the Green band in the image
            index_swir: index of the Short-wave Infrared (SWIR) band in the image
        """
        super().__init__(index_a=index_green, index_b=index_swir)


class AppendNDVI(AppendNormalizedDifferenceIndex):
    r"""Normalized Difference Vegetation Index (NDVI).

    Computes the following index:

    .. math::

       \text{NDVI} = \frac{\text{NIR} - \text{R}}{\text{NIR} + \text{R}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1016/0034-4257(79)90013-0
    """

    def __init__(self, index_nir: int, index_red: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the Near Infrared (NIR) band in the image
            index_red: index of the Red band in the image
        """
        super().__init__(index_a=index_nir, index_b=index_red)


class AppendNDWI(AppendNormalizedDifferenceIndex):
    r"""Normalized Difference Water Index (NDWI).

    Computes the following index:

    .. math::

       \text{NDWI} = \frac{\text{G} - \text{NIR}}{\text{G} + \text{NIR}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1080/01431169608948714
    """

    def __init__(self, index_green: int, index_nir: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_green: index of the Green band in the image
            index_nir: index of the Near Infrared (NIR) band in the image
        """
        super().__init__(index_a=index_green, index_b=index_nir)


class AppendSWI(AppendNormalizedDifferenceIndex):
    r"""Standardized Water-Level Index (SWI).

    Computes the following index:

    .. math::

       \text{SWI} = \frac{\text{VRE1} - \text{SWIR2}}{\text{VRE1} + \text{SWIR2}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.3390/w13121647

    .. versionadded:: 0.3
    """

    def __init__(self, index_vre1: int, index_swir2: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_vre1: index of the VRE1 band, e.g. B5 in Sentinel 2 imagery
            index_swir2: index of the SWIR2 band, e.g. B11 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_vre1, index_b=index_swir2)


class AppendGNDVI(AppendNormalizedDifferenceIndex):
    r"""Green Normalized Difference Vegetation Index (GNDVI).

    Computes the following index:

    .. math::

       \text{GNDVI} = \frac{\text{NIR} - \text{G}}{\text{NIR} + \text{G}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.2134/agronj2001.933583x

    .. versionadded:: 0.3
    """

    def __init__(self, index_nir: int, index_green: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the NIR band, e.g. B8 in Sentinel 2 imagery
            index_green: index of the Green band, e.g. B3 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_nir, index_b=index_green)


class AppendBNDVI(AppendNormalizedDifferenceIndex):
    r"""Blue Normalized Difference Vegetation Index (BNDVI).

    Computes the following index:

    .. math::

       \text{BNDVI} = \frac{\text{NIR} - \text{B}}{\text{NIR} + \text{B}}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1016/S1672-6308(07)60027-4

    .. versionadded:: 0.3
    """

    def __init__(self, index_nir: int, index_blue: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the NIR band, e.g. B8 in Sentinel 2 imagery
            index_blue: index of the Blue band, e.g. B2 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_nir, index_b=index_blue)


class AppendNDRE(AppendNormalizedDifferenceIndex):
    r"""Normalized Difference Red Edge Vegetation Index (NDRE).

    Computes the following index:

    .. math::

       \text{NDRE} = \frac{\text{NIR} - \text{VRE1}}{\text{NIR} + \text{VRE1}}

    If you use this index in your research, please cite the following paper:

    * https://agris.fao.org/agris-search/search.do?recordID=US201300795763

    .. versionadded:: 0.3
    """

    def __init__(self, index_nir: int, index_vre1: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the NIR band, e.g. B8 in Sentinel 2 imagery
            index_vre1: index of the Red Edge band, B5 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_nir, index_b=index_vre1)


class AppendTriBandNormalizedDifferenceIndex(IntensityAugmentationBase2D):
    r"""Append normalized difference index involving 3 bands as channel to image tensor.

    Computes the following index:

    .. math::

       \text{TBNDI} = \frac{A - (B + C)}{A + (B + C)}

    .. versionadded:: 0.3
    """

    def __init__(self, index_a: int, index_b: int, index_c: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_a: reference band channel index
            index_b: difference band channel index of component 1
            index_c: difference band channel index of component 2
        """
        super().__init__(p=1)
        self.flags = {'index_a': index_a, 'index_b': index_b, 'index_c': index_c}

    def apply_transform(
        self,
        input: Tensor,
        params: dict[str, Tensor],
        flags: dict[str, int],
        transform: Tensor | None = None,
    ) -> Tensor:
        """Apply the transform.

        Args:
            input: the input tensor
            params: generated parameters
            flags: static parameters
            transform: the geometric transformation tensor

        Returns:
            the augmented input
        """
        band_a = input[..., flags['index_a'], :, :]
        band_b = input[..., flags['index_b'], :, :]
        band_c = input[..., flags['index_c'], :, :]
        band_d = band_b + band_c
        tbndi = (band_a - band_d) / (band_a + band_d + _EPSILON)
        tbndi = torch.unsqueeze(tbndi, -3)
        input = torch.cat((input, tbndi), dim=-3)
        return input


class AppendGRNDVI(AppendTriBandNormalizedDifferenceIndex):
    r"""Green-Red Normalized Difference Vegetation Index (GRNDVI).

    Computes the following index:

    .. math::

       \text{GRNDVI} =
           \frac{\text{NIR} - (\text{G} + \text{R})}{\text{NIR} + (\text{G} + \text{R})}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1016/S1672-6308(07)60027-4

    .. versionadded:: 0.3
    """

    def __init__(self, index_nir: int, index_green: int, index_red: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the NIR band, e.g. B8 in Sentinel 2 imagery
            index_green: index of the Green band, B3 in Sentinel 2 imagery
            index_red: index of the Red band, B4 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_nir, index_b=index_green, index_c=index_red)


class AppendGBNDVI(AppendTriBandNormalizedDifferenceIndex):
    r"""Green-Blue Normalized Difference Vegetation Index (GBNDVI).

    Computes the following index:

    .. math::

       \text{GBNDVI} =
           \frac{\text{NIR} - (\text{G} + \text{B})}{\text{NIR} + (\text{G} + \text{B})}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1016/S1672-6308(07)60027-4

    .. versionadded:: 0.3
    """

    def __init__(self, index_nir: int, index_green: int, index_blue: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the NIR band, e.g. B8 in Sentinel 2 imagery
            index_green: index of the Green band, B3 in Sentinel 2 imagery
            index_blue: index of the Blue band, B2 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_nir, index_b=index_green, index_c=index_blue)


class AppendRBNDVI(AppendTriBandNormalizedDifferenceIndex):
    r"""Red-Blue Normalized Difference Vegetation Index (RBNDVI).

    Computes the following index:

    .. math::

       \text{RBNDVI} =
           \frac{\text{NIR} - (\text{R} + \text{B})}{\text{NIR} + (\text{R} + \text{B})}

    If you use this index in your research, please cite the following paper:

    * https://doi.org/10.1016/S1672-6308(07)60027-4

    .. versionadded:: 0.3
    """

    def __init__(self, index_nir: int, index_red: int, index_blue: int) -> None:
        """Initialize a new transform instance.

        Args:
            index_nir: index of the NIR band, e.g. B8 in Sentinel 2 imagery
            index_red: index of the Red band, B4 in Sentinel 2 imagery
            index_blue: index of the Blue band, B2 in Sentinel 2 imagery
        """
        super().__init__(index_a=index_nir, index_b=index_red, index_c=index_blue)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""MDAS dataset."""

import os
from collections.abc import Callable
from typing import Any, ClassVar

import matplotlib.pyplot as plt
import numpy as np
import rasterio as rio
import torch
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_and_extract_archive, extract_archive


class MDAS(NonGeoDataset):
    """MDAS dataset.

    The `MDAS <https://essd.copernicus.org/articles/15/113/2023/>`__ multimodal dataset
    is a comprehensive dataset for the city of Augsburg, Germany, collected on 7th May 2018.
    It includes SAR, multispectral, hyperspectral, DSM, and GIS data,
    providing comprehensive options for data fusion research.
    MDAS supports applications like resolution enhancement, spectral unmixing, and land cover classification.

    Dataset features:

    * 3K DSM data
    * 3K high resolution RGB images
    * Original very high resolution HySpex airborne imagery
    * EeteS simulated imagery with 10m GSD and EnMAP spectral bands
    * EeteS simulated imagery with 30m GSD and EnMAP spectral bands
    * EeteS simulated imagery with 10m GSD and Sentinel-2 spectral bands
    * Sentinel-2 L2A product
    * Sentinel-1 GRD product
    * Open Street Map (OSM) labels, see `this table <https://github.com/zhu-xlab/augsburg_Multimodal_Data_Set_MDaS/blob/75c015022b5f688dfc44744f19bcf34bdce786c7/Augsburg_data_4_publication/entire_city/OSM_label/README#L14>`__ for
      a table of the label distribution

    Dataset format:

    * 3K_RGB.tif (Shape: (4, 15000, 18000)px, Data Type: uint8)
    * 3K_dsm.tif (Shape: (1, 10000, 12000)px, Data Type: float32)
    * HySpex.tif (Shape: (368, 1364, 1636)px, Data Type: int16)
    * EeteS_EnMAP_2dot2m.tif (Shape: (242, 1364, 1636)px, Data Type: float32)
    * EeteS_EnMAP_10m.tif (Shape: (242, 300, 360)px, Data Type: uint16)
    * EeteS_EnMAP_30m.tif (Shape: (242, 100, 120)px, Data Type: uint16)
    * EeteS_Sentinel_2_10m.tif (Shape: (4, 300, 360)px, Data Type: uint16)
    * Sentinel_2.tif (Shape: (12, 300, 360)px, Data Type: uint16)
    * Sentinel_1.tif (Shape: (2, 300, 360)px, Data Type: float32)
    * osm_buildings.tif (Shape: (1, 1364, 1636)px, Data Type: uint8)
    * osm_landuse.tif (Shape: (1, 1364, 1636)px, Data Type: float64)
    * osm_water.tif (Shape: (1, 1364, 1636)px, Data Type: float64)

    If you use this dataset in your research, please cite the following paper:

    * https://essd.copernicus.org/articles/15/113/2023/

    .. versionadded:: 0.7
    """

    valid_modalities = (
        '3K_DSM',
        '3K_RGB',
        'HySpex',
        'EeteS_EnMAP_10m',
        'EeteS_EnMAP_30m',
        'EeteS_Sentinel_2_10m',
        'Sentinel_2',
        'Sentinel_1',
        'osm_buildings',
        'osm_landuse',
        'osm_water',
    )
    landuse_class_names: ClassVar[dict[int, str]] = {
        0: 'no label',
        1: 'forest',
        2: 'park',
        3: 'residential',
        4: 'industrial',
        5: 'farm',
        6: 'cemetery',
        7: 'allotments',
        8: 'meadow',
        9: 'commercial',
        10: 'nature reserve',
        11: 'recreation ground',
        12: 'retail',
        13: 'military',
        14: 'quarry',
        15: 'orchard',
        16: 'scrub',
        17: 'grass',
        18: 'heath',
    }

    # https://github.com/zhu-xlab/augsburg_Multimodal_Data_Set_MDaS/blob/75c015022b5f688dfc44744f19bcf34bdce786c7/Augsburg_data_4_publication/entire_city/OSM_label/README#L14
    landuse_mapping: ClassVar[dict[int, int]] = {
        -2147483647: 0,
        7201: 1,
        7202: 2,
        7203: 3,
        7204: 4,
        7205: 5,
        7206: 6,
        7207: 7,
        7208: 8,
        7209: 9,
        7210: 10,
        7211: 11,
        7212: 12,
        7213: 13,
        7214: 14,
        7215: 15,
        7217: 16,
        7218: 17,
        7219: 18,
    }

    ds_root_name = 'Augsburg_data_4_publication'

    zipfilename = f'{ds_root_name}.zip'

    valid_subareas = ('sub_area_1', 'sub_area_2', 'sub_area_3')

    url = 'https://huggingface.co/datasets/torchgeo/mdas/resolve/860226b74269f1cf1bed8ea3c03f571ae701144c/Augsburg_data_4_publication.zip'

    md5 = '7b63c26e3717cb52c6ba47d215f18d5b'

    enmap_rgb_band_idx: ClassVar[list[int]] = [43, 28, 10]
    sentinel_2_rgb_band_idx: ClassVar[list[int]] = [3, 2, 1]
    hyspex_rgb_band_idx: ClassVar[list[int]] = [100, 50, 10]

    def __init__(
        self,
        root: Path = 'data',
        subareas: list[str] = ['sub_area_1'],
        modalities: list[str] = ['3K_RGB', 'HySpex', 'Sentinel_2'],
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new MDAS dataset instance.

        Args:
            root: Root directory where the dataset should be stored.
            subareas: The subareas to load. Options are 'sub_area_1', 'sub_area_2', 'sub_area_3'.
            modalities: The modalities to load. Options are '3K_DSM', '3K_RGB', 'HySpex', 'EeteS_EnMAP_10m', 'EeteS_EnMAP_30m', 'EeteS_Sentinel_2_10m', 'Sentinel-2', 'Sentinel-1', 'OSM_label'.
            transforms: A function/transform that takes in a dictionary and returns a transformed version.
            download: if True, download dataset and store it in the root directory
            checksum: If True, check the integrity of the dataset after download.

        Raises:
            AssertionError: If the subareas or modalities are not valid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.root = root
        self.download = download
        assert all(sub in self.valid_subareas for sub in subareas), (
            f'Subareas must be one of {self.valid_subareas}'
        )
        self.subareas = subareas
        assert all(mod in self.valid_modalities for mod in modalities), (
            f'Modalities must be one of {self.valid_modalities}'
        )
        self.modalities = modalities
        self.transforms = transforms
        self.checksum = checksum

        self._verify()
        self.files = self._load_files()

    def __len__(self) -> int:
        """Return the number of samples in the dataset.

        Returns:
            the length of the dataset
        """
        return len(self.files)

    def _load_files(self) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Returns:
            a list of dictionaries containing the paths of the files in the dataset
        """
        files = []
        for subarea in self.subareas:
            subarea_files = {}
            for modality in self.modalities:
                subarea_files[modality] = os.path.join(
                    self.root,
                    self.ds_root_name,
                    subarea,
                    f'{modality}_{self._format_subarea(subarea)}.tif',
                )
            files.append(subarea_files)
        return files

    def _format_subarea(self, subarea: str) -> str:
        """Format the subarea name.

        Args:
            subarea: The subarea string to format.

        Returns:
            formatted subarea string for files
        """
        parts = subarea.split('_')
        return parts[0] + '_' + parts[1] + parts[2]

    def _load_image(self, path: Path) -> Tensor:
        """Load an image from a given path.

        Args:
            path: The path to the image file

        Returns:
            the loaded image as a tensor
        """
        with rio.open(path) as src:
            img = src.read()
            if img.dtype == np.uint16:
                img = img.astype(np.int32)
            if 'osm_landuse' in str(path):
                img = np.vectorize(self.landuse_mapping.get)(img)

            return torch.from_numpy(img)

    def __getitem__(self, idx: int) -> dict[str, Tensor]:
        """Return the dataset sample at the given index.

        Args:
            idx: The index of the sample to return

        Returns:
            a dictionary containing the data of chosen modalities
        """
        sample_files = self.files[idx]
        sample: dict[str, Any] = {}
        for modality, path in sample_files.items():
            if 'osm' in modality:
                sample[f'{modality}_mask'] = self._load_image(path).long()
            else:
                sample[f'{modality}_image'] = self._load_image(path)

        if self.transforms:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # check if each desired modality file exists in specified subarea
        exists = []
        for subarea in self.subareas:
            for modality in self.modalities:
                path = os.path.join(
                    self.root,
                    self.ds_root_name,
                    subarea,
                    f'{modality}_{self._format_subarea(subarea)}.tif',
                )
                if not os.path.exists(path):
                    exists.append(False)
                else:
                    exists.append(True)
        if all(exists):
            return

        # check if zip file downloaded
        if os.path.exists(os.path.join(self.root, self.zipfilename)):
            self._extract()
            return

        if not self.download:
            raise DatasetNotFoundError(self)

        self._download()

    def _extract(self) -> None:
        """Extract the dataset."""
        extract_archive(os.path.join(self.root, self.zipfilename), self.root)

    def _download(self) -> None:
        """Download the dataset."""
        download_and_extract_archive(
            self.url,
            self.root,
            filename=self.zipfilename,
            md5=self.md5 if self.checksum else None,
        )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: A sample returned by `__getitem__`.
            show_titles: Whether to display titles on the subplots.
            suptitle: An optional super title for the plot.

        Returns:
            a matplotlib Figure with the rendered sample
        """
        ncols = len(sample)
        fig, axs = plt.subplots(1, ncols, figsize=(5 * ncols, 5))

        if ncols == 1:
            axs = [axs]

        for idx, (key, data) in enumerate(sample.items()):
            match key:
                case '3K_RGB_image':
                    img = data[:3].numpy().transpose(1, 2, 0) / 255.0
                    axs[idx].imshow(img)
                case '3K_DSM_image':
                    img = data.numpy().squeeze(0)
                    axs[idx].imshow(img, cmap='gray')
                case 'EeteS_EnMAP_10m_image' | 'EeteS_EnMAP_30m_image':
                    img = (
                        data[self.enmap_rgb_band_idx].numpy().transpose(1, 2, 0)
                        / 10000.0
                    )
                    axs[idx].imshow(img)
                case 'EeteS_Sentinel_2_10m_image':
                    img = (
                        data[self.sentinel_2_rgb_band_idx].numpy().transpose(1, 2, 0)
                        / 10000.0
                    )
                    axs[idx].imshow(img)
                case 'Sentinel_1_image':
                    img = data[0].numpy().clip(0, 1)
                    axs[idx].imshow(img)
                case 'Sentinel_2_image':
                    img = (
                        data[self.sentinel_2_rgb_band_idx].numpy().transpose(1, 2, 0)
                        / 10000.0
                    )
                    axs[idx].imshow(img)
                case 'HySpex_image':
                    img = (
                        data[self.hyspex_rgb_band_idx].numpy().transpose(1, 2, 0)
                        / 15000.0
                    )
                    axs[idx].imshow(img)
                case 'osm_landuse_mask':
                    img = data.numpy().squeeze(0)
                    cmap = ListedColormap([plt.get_cmap('tab20')(i) for i in range(20)])
                    im = axs[idx].imshow(img, cmap=cmap)
                    cbar = plt.colorbar(im, ax=axs[idx], ticks=range(19))
                    cbar.ax.set_yticklabels(
                        [self.landuse_class_names[i] for i in range(19)]
                    )
                case 'osm_buildings_mask':
                    img = data.numpy().squeeze(0)
                    axs[idx].imshow(img, cmap='gray')
                case 'osm_water_mask':
                    img = data.numpy().squeeze(0)
                    axs[idx].imshow(img, cmap='Blues')

            axs[idx].axis('off')
            if show_titles:
                axs[idx].set_title(key)

        if suptitle:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""ADVANCE dataset."""

import glob
import os
from collections.abc import Callable
from typing import cast

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_and_extract_archive, lazy_import


class ADVANCE(NonGeoDataset):
    """ADVANCE dataset.

    The `ADVANCE <https://akchen.github.io/ADVANCE-DATASET/>`__
    dataset is a dataset for audio visual scene recognition.

    Dataset features:

    * 5,075 pairs of geotagged audio recordings and images
    * three spectral bands - RGB (512x512 px)
    * 10-second audio recordings

    Dataset format:

    * images are three-channel jpgs
    * audio files are in wav format

    Dataset classes:

    0. airport
    1. beach
    2. bridge
    3. farmland
    4. forest
    5. grassland
    6. harbour
    7. lake
    8. orchard
    9. residential
    10. sparse shrub land
    11. sports land
    12. train station

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1007/978-3-030-58586-0_5

    .. note::
        This dataset requires the following additional library to be installed:

        * `scipy <https://pypi.org/project/scipy/>`_ to load the audio files to tensors
    """

    urls = (
        'https://zenodo.org/records/3828124/files/ADVANCE_vision.zip?download=1',
        'https://zenodo.org/records/3828124/files/ADVANCE_sound.zip?download=1',
    )
    filenames = ('ADVANCE_vision.zip', 'ADVANCE_sound.zip')
    md5s = ('a9e8748219ef5864d3b5a8979a67b471', 'a2d12f2d2a64f5c3d3a9d8c09aaf1c31')
    directories = ('vision', 'sound')
    classes: tuple[str, ...] = (
        'airport',
        'beach',
        'bridge',
        'farmland',
        'forest',
        'grassland',
        'harbour',
        'lake',
        'orchard',
        'residential',
        'sparse shrub land',
        'sports land',
        'train station',
    )

    def __init__(
        self,
        root: Path = 'data',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new ADVANCE dataset instance.

        Args:
            root: root directory where dataset can be found
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If scipy is not installed.
        """
        lazy_import('scipy.io.wavfile')

        self.root = root
        self.transforms = transforms
        self.checksum = checksum

        if download:
            self._download()

        if not self._check_integrity():
            raise DatasetNotFoundError(self)

        self.files = self._load_files(self.root)
        self.classes = tuple(sorted({f['cls'] for f in self.files}))
        self.class_to_idx: dict[str, int] = {c: i for i, c in enumerate(self.classes)}

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        files = self.files[index]
        image = self._load_image(files['image'])
        audio = self._load_target(files['audio'])
        cls_label = self.class_to_idx[files['cls']]
        label = torch.tensor(cls_label, dtype=torch.long)
        sample = {'image': image, 'audio': audio, 'label': label}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def _load_files(self, root: Path) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Args:
            root: root dir of dataset

        Returns:
            list of dicts containing paths for each pair of image, audio, label
        """
        images = sorted(glob.glob(os.path.join(root, 'vision', '**', '*.jpg')))
        wavs = sorted(glob.glob(os.path.join(root, 'sound', '**', '*.wav')))
        labels = [image.split(os.sep)[-2] for image in images]
        files = [
            dict(image=image, audio=wav, cls=label)
            for image, wav, label in zip(images, wavs, labels)
        ]
        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the image
        """
        with Image.open(path) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('RGB'))
            tensor = torch.from_numpy(array)
            # Convert from HxWxC to CxHxW
            tensor = tensor.permute((2, 0, 1))
            return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Load the target audio for a single image.

        Args:
            path: path to the target

        Returns:
            the target audio
        """
        siw = lazy_import('scipy.io.wavfile')
        array = siw.read(path, mmap=True)[1]
        tensor = torch.from_numpy(array)
        tensor = tensor.unsqueeze(0)
        return tensor

    def _check_integrity(self) -> bool:
        """Checks the integrity of the dataset structure.

        Returns:
            True if the dataset directories are found, else False
        """
        for directory in self.directories:
            filepath = os.path.join(self.root, directory)
            if not os.path.exists(filepath):
                return False
        return True

    def _download(self) -> None:
        """Download the dataset and extract it."""
        if self._check_integrity():
            print('Files already downloaded and verified')
            return

        for filename, url, md5 in zip(self.filenames, self.urls, self.md5s):
            download_and_extract_archive(
                url, self.root, filename=filename, md5=md5 if self.checksum else None
            )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        image = np.rollaxis(sample['image'].numpy(), 0, 3)
        label = cast(int, sample['label'].item())
        label_class = self.classes[label]

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = cast(int, sample['prediction'].item())
            prediction_class = self.classes[prediction]

        fig, ax = plt.subplots(figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')
        if show_titles:
            title = f'Label: {label_class}'
            if showing_predictions:
                title += f'\nPrediction: {prediction_class}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TorchGeo datasets."""

from .advance import ADVANCE
from .agb_live_woody_density import AbovegroundLiveWoodyBiomassDensity
from .agrifieldnet import AgriFieldNet
from .airphen import Airphen
from .astergdem import AsterGDEM
from .benin_cashews import BeninSmallHolderCashews
from .bigearthnet import BigEarthNet, BigEarthNetV2
from .biomassters import BioMassters
from .bright import BRIGHTDFC2025
from .cabuar import CaBuAr
from .caffe import CaFFe
from .cbf import CanadianBuildingFootprints
from .cdl import CDL
from .chabud import ChaBuD
from .chesapeake import (
    Chesapeake,
    ChesapeakeCVPR,
    ChesapeakeDC,
    ChesapeakeDE,
    ChesapeakeMD,
    ChesapeakeNY,
    ChesapeakePA,
    ChesapeakeVA,
    ChesapeakeWV,
)
from .cloud_cover import CloudCoverDetection
from .cms_mangrove_canopy import CMSGlobalMangroveCanopy
from .copernicus import (
    CopernicusBench,
    CopernicusBenchAQNO2S5P,
    CopernicusBenchAQO3S5P,
    CopernicusBenchBase,
    CopernicusBenchBigEarthNetS1,
    CopernicusBenchBigEarthNetS2,
    CopernicusBenchBiomassS3,
    CopernicusBenchCloudS2,
    CopernicusBenchCloudS3,
    CopernicusBenchDFC2020S1,
    CopernicusBenchDFC2020S2,
    CopernicusBenchEuroSATS1,
    CopernicusBenchEuroSATS2,
    CopernicusBenchFloodS1,
    CopernicusBenchLC100ClsS3,
    CopernicusBenchLC100SegS3,
    CopernicusBenchLCZS2,
    CopernicusPretrain,
)
from .cowc import COWC, COWCCounting, COWCDetection
from .cropharvest import CropHarvest
from .cv4a_kenya_crop_type import CV4AKenyaCropType
from .cyclone import TropicalCyclone
from .deepglobelandcover import DeepGlobeLandCover
from .dfc2022 import DFC2022
from .digital_typhoon import DigitalTyphoon
from .dior import DIOR
from .dl4gam import DL4GAMAlps
from .dota import DOTA
from .eddmaps import EDDMapS
from .enmap import EnMAP
from .enviroatlas import EnviroAtlas
from .errors import DatasetNotFoundError, DependencyNotFoundError, RGBBandsMissingError
from .esri2020 import Esri2020
from .etci2021 import ETCI2021
from .eudem import EUDEM
from .eurocrops import EuroCrops
from .eurosat import EuroSAT, EuroSAT100, EuroSATSpatial
from .everwatch import EverWatch
from .fair1m import FAIR1M
from .fire_risk import FireRisk
from .forestdamage import ForestDamage
from .ftw import FieldsOfTheWorld
from .gbif import GBIF
from .gbm import GlobalBuildingMap
from .geo import (
    GeoDataset,
    IntersectionDataset,
    NonGeoClassificationDataset,
    NonGeoDataset,
    RasterDataset,
    UnionDataset,
    VectorDataset,
)
from .geonrw import GeoNRW
from .gid15 import GID15
from .globbiomass import GlobBiomass
from .hyspecnet import HySpecNet11k
from .idtrees import IDTReeS
from .inaturalist import INaturalist
from .inria import InriaAerialImageLabeling
from .iobench import IOBench
from .l7irish import L7Irish
from .l8biome import L8Biome
from .landcoverai import LandCoverAI, LandCoverAI100, LandCoverAIBase, LandCoverAIGeo
from .landsat import (
    Landsat,
    Landsat1,
    Landsat2,
    Landsat3,
    Landsat4MSS,
    Landsat4TM,
    Landsat5MSS,
    Landsat5TM,
    Landsat7,
    Landsat8,
    Landsat9,
)
from .levircd import LEVIRCD, LEVIRCDBase, LEVIRCDPlus
from .loveda import LoveDA
from .mapinwild import MapInWild
from .mdas import MDAS
from .millionaid import MillionAID
from .mmearth import MMEarth
from .mmflood import MMFlood
from .naip import NAIP
from .nasa_marine_debris import NASAMarineDebris
from .nccm import NCCM
from .nlcd import NLCD
from .openbuildings import OpenBuildings
from .oscd import OSCD
from .pastis import PASTIS
from .patternnet import PatternNet
from .potsdam import Potsdam2D
from .prisma import PRISMA
from .quakeset import QuakeSet
from .reforestree import ReforesTree
from .resisc45 import RESISC45
from .rwanda_field_boundary import RwandaFieldBoundary
from .satlas import SatlasPretrain
from .seasonet import SeasoNet
from .seco import SeasonalContrastS2
from .sen12ms import SEN12MS
from .sentinel import Sentinel, Sentinel1, Sentinel2
from .skippd import SKIPPD
from .skyscript import SkyScript
from .so2sat import So2Sat
from .soda import SODAA
from .south_africa_crop_type import SouthAfricaCropType
from .south_america_soybean import SouthAmericaSoybean
from .spacenet import (
    SpaceNet,
    SpaceNet1,
    SpaceNet2,
    SpaceNet3,
    SpaceNet4,
    SpaceNet5,
    SpaceNet6,
    SpaceNet7,
    SpaceNet8,
)
from .splits import (
    random_bbox_assignment,
    random_bbox_splitting,
    random_grid_cell_assignment,
    roi_split,
    time_series_split,
)
from .ssl4eo import SSL4EO, SSL4EOL, SSL4EOS12
from .ssl4eo_benchmark import SSL4EOLBenchmark
from .substation import Substation
from .sustainbench_crop_yield import SustainBenchCropYield
from .treesatai import TreeSatAI
from .ucmerced import UCMerced
from .usavars import USAVars
from .utils import (
    BoundingBox,
    concat_samples,
    merge_samples,
    stack_samples,
    unbind_samples,
)
from .vaihingen import Vaihingen2D
from .vhr10 import VHR10
from .western_usa_live_fuel_moisture import WesternUSALiveFuelMoisture
from .xview import XView2
from .zuericrop import ZueriCrop

__all__ = (
    'ADVANCE',
    'BRIGHTDFC2025',
    'CDL',
    'COWC',
    'DFC2022',
    'DIOR',
    'DOTA',
    'ETCI2021',
    'EUDEM',
    'FAIR1M',
    'GBIF',
    'GID15',
    'LEVIRCD',
    'MDAS',
    'NAIP',
    'NCCM',
    'NLCD',
    'OSCD',
    'PASTIS',
    'PRISMA',
    'RESISC45',
    'SEN12MS',
    'SKIPPD',
    'SODAA',
    'SSL4EO',
    'SSL4EOL',
    'SSL4EOS12',
    'VHR10',
    'AbovegroundLiveWoodyBiomassDensity',
    'AgriFieldNet',
    'Airphen',
    'AsterGDEM',
    'BeninSmallHolderCashews',
    'BigEarthNet',
    'BigEarthNetV2',
    'BioMassters',
    'BoundingBox',
    'CMSGlobalMangroveCanopy',
    'COWCCounting',
    'COWCDetection',
    'CV4AKenyaCropType',
    'CaBuAr',
    'CaFFe',
    'CanadianBuildingFootprints',
    'ChaBuD',
    'Chesapeake',
    'ChesapeakeCVPR',
    'ChesapeakeDC',
    'ChesapeakeDE',
    'ChesapeakeMD',
    'ChesapeakeNY',
    'ChesapeakePA',
    'ChesapeakeVA',
    'ChesapeakeWV',
    'CloudCoverDetection',
    'CopernicusBench',
    'CopernicusBenchAQNO2S5P',
    'CopernicusBenchAQO3S5P',
    'CopernicusBenchBase',
    'CopernicusBenchBigEarthNetS1',
    'CopernicusBenchBigEarthNetS2',
    'CopernicusBenchBiomassS3',
    'CopernicusBenchCloudS2',
    'CopernicusBenchCloudS3',
    'CopernicusBenchDFC2020S1',
    'CopernicusBenchDFC2020S2',
    'CopernicusBenchEuroSATS1',
    'CopernicusBenchEuroSATS2',
    'CopernicusBenchFloodS1',
    'CopernicusBenchLC100ClsS3',
    'CopernicusBenchLC100SegS3',
    'CopernicusBenchLCZS2',
    'CopernicusPretrain',
    'CropHarvest',
    'DL4GAMAlps',
    'DatasetNotFoundError',
    'DeepGlobeLandCover',
    'DependencyNotFoundError',
    'DigitalTyphoon',
    'EDDMapS',
    'EnMAP',
    'EnviroAtlas',
    'Esri2020',
    'EuroCrops',
    'EuroSAT',
    'EuroSAT100',
    'EuroSATSpatial',
    'EverWatch',
    'FieldsOfTheWorld',
    'FireRisk',
    'ForestDamage',
    'GeoDataset',
    'GeoNRW',
    'GlobBiomass',
    'GlobalBuildingMap',
    'HySpecNet11k',
    'IDTReeS',
    'INaturalist',
    'IOBench',
    'InriaAerialImageLabeling',
    'IntersectionDataset',
    'L7Irish',
    'L8Biome',
    'LEVIRCDBase',
    'LEVIRCDPlus',
    'LandCoverAI',
    'LandCoverAI100',
    'LandCoverAIBase',
    'LandCoverAIGeo',
    'Landsat',
    'Landsat1',
    'Landsat2',
    'Landsat3',
    'Landsat4MSS',
    'Landsat4TM',
    'Landsat5MSS',
    'Landsat5TM',
    'Landsat7',
    'Landsat8',
    'Landsat9',
    'LoveDA',
    'MMEarth',
    'MMFlood',
    'MapInWild',
    'MillionAID',
    'NASAMarineDebris',
    'NonGeoClassificationDataset',
    'NonGeoDataset',
    'OpenBuildings',
    'PatternNet',
    'Potsdam2D',
    'QuakeSet',
    'RGBBandsMissingError',
    'RasterDataset',
    'ReforesTree',
    'RwandaFieldBoundary',
    'SSL4EOLBenchmark',
    'SatlasPretrain',
    'SeasoNet',
    'SeasonalContrastS2',
    'Sentinel',
    'Sentinel1',
    'Sentinel2',
    'SkyScript',
    'So2Sat',
    'SouthAfricaCropType',
    'SouthAmericaSoybean',
    'SpaceNet',
    'SpaceNet1',
    'SpaceNet2',
    'SpaceNet3',
    'SpaceNet4',
    'SpaceNet5',
    'SpaceNet6',
    'SpaceNet7',
    'SpaceNet8',
    'Substation',
    'SustainBenchCropYield',
    'TreeSatAI',
    'TropicalCyclone',
    'UCMerced',
    'USAVars',
    'UnionDataset',
    'Vaihingen2D',
    'VectorDataset',
    'WesternUSALiveFuelMoisture',
    'XView2',
    'ZueriCrop',
    'concat_samples',
    'merge_samples',
    'random_bbox_assignment',
    'random_bbox_splitting',
    'random_grid_cell_assignment',
    'roi_split',
    'stack_samples',
    'time_series_split',
    'unbind_samples',
)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SEN12MS datamodule."""

from typing import Any

import torch
from torch import Tensor
from torch.utils.data import Subset

from ..datasets import SEN12MS
from .geo import NonGeoDataModule
from .utils import group_shuffle_split


class SEN12MSDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the SEN12MS dataset.

    Implements 80/20 geographic train/val splits and uses the test split from the
    classification dataset definitions.

    Uses the Simplified IGBP scheme defined in the 2020 Data Fusion Competition. See
    https://arxiv.org/abs/2002.08254.
    """

    #: Mapping from the IGBP class definitions to the DFC2020, taken from the dataloader
    #: here: https://github.com/lukasliebel/dfc2020_baseline.
    DFC2020_CLASS_MAPPING = torch.tensor(
        [0, 1, 1, 1, 1, 1, 2, 2, 3, 3, 4, 5, 6, 7, 6, 8, 9, 10]
    )

    std = torch.tensor(
        [-25, -25, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4, 1e4]
    )

    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 0,
        band_set: str = 'all',
        **kwargs: Any,
    ) -> None:
        """Initialize a new SEN12MSDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            band_set: Subset of S1/S2 bands to use. Options are: "all",
                "s1", "s2-all", and "s2-reduced" where the "s2-reduced" set includes:
                B2, B3, B4, B8, B11, and B12.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.SEN12MS`.
        """
        kwargs['bands'] = SEN12MS.BAND_SETS[band_set]

        if band_set == 's1':
            self.std = self.std[:2]
        elif band_set == 's2-all':
            self.std = self.std[2:]
        elif band_set == 's2-reduced':
            self.std = self.std[torch.tensor([3, 4, 5, 9, 12, 13])]

        super().__init__(SEN12MS, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit', 'validate']:
            season_to_int = {'winter': 0, 'spring': 1000, 'summer': 2000, 'fall': 3000}

            self.dataset = SEN12MS(split='train', **self.kwargs)

            # A patch is a filename like:
            #     "ROIs{num}_{season}_s2_{scene_id}_p{patch_id}.tif"
            # This patch will belong to the scene that is uniquely identified by its
            # (season, scene_id) tuple. Because the largest scene_id is 149, we can
            # simply give each season a large number and representing a unique_scene_id
            # as (season_id + scene_id).
            scenes = []
            for scene_fn in self.dataset.ids:
                parts = scene_fn.split('_')
                season_id = season_to_int[parts[1]]
                scene_id = int(parts[3])
                scenes.append(season_id + scene_id)

            train_indices, val_indices = group_shuffle_split(
                scenes, test_size=0.2, random_state=0
            )

            self.train_dataset = Subset(self.dataset, train_indices)
            self.val_dataset = Subset(self.dataset, val_indices)
        if stage in ['test']:
            self.test_dataset = SEN12MS(split='test', **self.kwargs)

    def on_after_batch_transfer(
        self, batch: dict[str, Tensor], dataloader_idx: int
    ) -> dict[str, Tensor]:
        """Apply batch augmentations to the batch after it is transferred to the device.

        Args:
            batch: A batch of data that needs to be altered or augmented.
            dataloader_idx: The index of the dataloader to which the batch belongs.

        Returns:
            A batch of data.
        """
        batch['mask'] = torch.take(self.DFC2020_CLASS_MAPPING, batch['mask'])

        return super().on_after_batch_transfer(batch, dataloader_idx)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 8 dataset."""

from typing import ClassVar

from .base import SpaceNet


class SpaceNet8(SpaceNet):
    r"""SpaceNet8: Flood Detection Challenge Using Multiclass Segmentation.

    `SpaceNet 8 <https://spacenet.ai/sn8-challenge/>`_ is a dataset focusing on
    infrastructure and flood mapping related to hurricanes and heavy rains that cause
    route obstructions and significant damage.

    If you use this dataset in your research, please cite the following paper:

    * https://openaccess.thecvf.com/content/CVPR2022W/EarthVision/html/Hansch_SpaceNet_8\_-_The_Detection_of_Flooded_Roads_and_Buildings_CVPRW_2022_paper.html

    .. versionadded:: 0.6
    """

    directory_glob = '{product}'
    file_regex = r'(\d+_\d+_\d+)\.'
    dataset_id = 'SN8_floods'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            0: [
                'Germany_Training_Public.tar.gz',
                'Louisiana-East_Training_Public.tar.gz',
            ]
        },
        'test': {0: ['Louisiana-West_Test_Public.tar.gz']},
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            0: ['81383a9050b93e8f70c8557d4568e8a2', 'fa40ae3cf6ac212c90073bf93d70bd95']
        },
        'test': {0: ['d41d8cd98f00b204e9800998ecf8427e']},
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {'train': [0], 'test': [0]}
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['PRE-event', 'POST-event'],
        'test': ['PRE-event', 'POST-event'],
    }
    valid_masks = ('annotations',)
    chip_size: ClassVar[dict[str, tuple[int, int]]] = {
        'PRE-event': (1300, 1300),
        'POST-event': (1300, 1300),
    }



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""LoveDA dataset."""

import glob
import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_and_extract_archive


class LoveDA(NonGeoDataset):
    """LoveDA dataset.

    The `LoveDA <https://github.com/Junjue-Wang/LoveDA>`__ datataset is a
    semantic segmentation dataset.

    Dataset features:

    * 2713 urban scene and 3274 rural scene HSR images, spatial resolution of 0.3m
    * image source is Google Earth platform
    * total of 166768 annotated objects from Nanjing, Changzhou and Wuhan cities
    * dataset comes with predefined train, validation, and test set
    * dataset differentiates between 'rural' and 'urban' images

    Dataset format:

    * images are three-channel pngs with dimension 1024x1024
    * segmentation masks are single-channel pngs

    Dataset classes:

    1. background
    2. building
    3. road
    4. water
    5. barren
    6. forest
    7. agriculture

    No-data regions assigned with 0 and should be ignored.

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2110.08733

    .. versionadded:: 0.2
    """

    scenes = ('urban', 'rural')
    splits = ('train', 'val', 'test')

    info_dict: ClassVar[dict[str, dict[str, str]]] = {
        'train': {
            'url': 'https://zenodo.org/records/5706578/files/Train.zip?download=1',
            'filename': 'Train.zip',
            'md5': 'de2b196043ed9b4af1690b3f9a7d558f',
        },
        'val': {
            'url': 'https://zenodo.org/records/5706578/files/Val.zip?download=1',
            'filename': 'Val.zip',
            'md5': '84cae2577468ff0b5386758bb386d31d',
        },
        'test': {
            'url': 'https://zenodo.org/records/5706578/files/Test.zip?download=1',
            'filename': 'Test.zip',
            'md5': 'a489be0090465e01fb067795d24e6b47',
        },
    }

    classes = (
        'background',
        'building',
        'road',
        'water',
        'barren',
        'forest',
        'agriculture',
        'no-data',
    )

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        scene: Sequence[str] = ['urban', 'rural'],
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new LoveDA dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            scene: specify whether to load only 'urban', only 'rural' or both
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` or ``scene`` arguments are invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        assert set(scene).intersection(set(self.scenes)), (
            "The possible scenes are 'rural' and/or 'urban'"
        )
        assert len(scene) <= 2, "There are no other scenes than 'rural' or 'urban'"

        self.root = root
        self.split = split
        self.scene = scene
        self.transforms = transforms
        self.checksum = checksum

        self.url = self.info_dict[self.split]['url']
        self.filename = self.info_dict[self.split]['filename']
        self.md5 = self.info_dict[self.split]['md5']

        self.directory = os.path.join(self.root, split.capitalize())
        self.scene_paths = [
            os.path.join(self.directory, s.capitalize()) for s in self.scene
        ]

        if download:
            self._download()

        if not self._check_integrity():
            raise DatasetNotFoundError(self)

        self.files = self._load_files(self.scene_paths, self.split)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            image and mask at that index with image of dimension 3x1024x1024
            and mask of dimension 1024x1024
        """
        files = self.files[index]
        image = self._load_image(files['image'])

        if self.split != 'test':
            mask = self._load_target(files['mask'])
            sample = {'image': image, 'mask': mask}
        else:
            sample = {'image': image}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of datapoints in the dataset.

        Returns:
            length of dataset
        """
        return len(self.files)

    def _load_files(self, scene_paths: list[str], split: str) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Args:
            scene_paths: contains one or two paths, depending on whether user has
                         specified only 'rural', 'only 'urban' or both
            split: subset of dataset, one of [train, val, test]
        """
        images = []

        for s in scene_paths:
            images.extend(glob.glob(os.path.join(s, 'images_png', '*.png')))

        images = sorted(images)

        if self.split != 'test':
            masks = [image.replace('images_png', 'masks_png') for image in images]
            files = [dict(image=image, mask=mask) for image, mask in zip(images, masks)]
        else:
            files = [dict(image=image) for image in images]

        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the loaded image
        """
        filename = os.path.join(path)
        with Image.open(filename) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('RGB'))
            tensor = torch.from_numpy(array).float()
            # Convert from HxWxC to CxHxW
            tensor = tensor.permute((2, 0, 1))
            return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Load a single mask corresponding to image.

        Args:
            path: path to the mask

        Returns:
            the mask of the image
        """
        filename = os.path.join(path)
        with Image.open(filename) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('L'))
            tensor = torch.from_numpy(array)
            tensor = tensor.to(torch.long)
            return tensor

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset structure.

        Returns:
            True if the dataset directories and split files are found, else False
        """
        for s in self.scene_paths:
            if not os.path.exists(s):
                return False

        return True

    def _download(self) -> None:
        """Download the dataset and extract it."""
        if self._check_integrity():
            print('Files already downloaded and verified')
            return

        download_and_extract_archive(
            self.url,
            self.root,
            filename=self.filename,
            md5=self.md5 if self.checksum else None,
        )

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample return by :meth:`__getitem__`
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        if self.split != 'test':
            image, mask = sample['image'], sample['mask']
            ncols = 2
        else:
            image = sample['image']
            ncols = 1

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(10, ncols * 10))

        if self.split != 'test':
            axs[0].imshow(image.permute(1, 2, 0))
            axs[0].axis('off')
            axs[1].imshow(mask)
            axs[1].axis('off')
        else:
            axs.imshow(image.permute(1, 2, 0))
            axs.axis('off')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig
